# Generated by ariadne-codegen
# Source: openhexa/graphql/schema.generated.graphql

from enum import Enum


class AccessmodAccessRequestStatus(str, Enum):
    APPROVED = "APPROVED"
    DENIED = "DENIED"
    PENDING = "PENDING"


class AccessmodAccessibilityAnalysisAlgorithm(str, Enum):
    ANISOTROPIC = "ANISOTROPIC"
    ISOTROPIC = "ISOTROPIC"


class AccessmodAnalysisStatus(str, Enum):
    DRAFT = "DRAFT"
    FAILED = "FAILED"
    QUEUED = "QUEUED"
    READY = "READY"
    RUNNING = "RUNNING"
    SUCCESS = "SUCCESS"


class AccessmodAnalysisType(str, Enum):
    ACCESSIBILITY = "ACCESSIBILITY"
    GEOGRAPHIC_COVERAGE = "GEOGRAPHIC_COVERAGE"
    ZONAL_STATISTICS = "ZONAL_STATISTICS"


class AccessmodFilesetFormat(str, Enum):
    RASTER = "RASTER"
    TABULAR = "TABULAR"
    VECTOR = "VECTOR"


class AccessmodFilesetMode(str, Enum):
    AUTOMATIC_ACQUISITION = "AUTOMATIC_ACQUISITION"
    USER_INPUT = "USER_INPUT"


class AccessmodFilesetRoleCode(str, Enum):
    BARRIER = "BARRIER"
    BOUNDARIES = "BOUNDARIES"
    COVERAGE = "COVERAGE"
    DEM = "DEM"
    FRICTION_SURFACE = "FRICTION_SURFACE"
    GEOMETRY = "GEOMETRY"
    HEALTH_FACILITIES = "HEALTH_FACILITIES"
    LAND_COVER = "LAND_COVER"
    POPULATION = "POPULATION"
    STACK = "STACK"
    TRANSPORT_NETWORK = "TRANSPORT_NETWORK"
    TRAVEL_TIMES = "TRAVEL_TIMES"
    WATER = "WATER"
    ZONAL_STATISTICS = "ZONAL_STATISTICS"
    ZONAL_STATISTICS_TABLE = "ZONAL_STATISTICS_TABLE"


class AccessmodFilesetStatus(str, Enum):
    INVALID = "INVALID"
    PENDING = "PENDING"
    TO_ACQUIRE = "TO_ACQUIRE"
    VALID = "VALID"
    VALIDATING = "VALIDATING"


class AccessmodProjectOrder(str, Enum):
    NAME_ASC = "NAME_ASC"
    NAME_DESC = "NAME_DESC"
    UPDATED_AT_ASC = "UPDATED_AT_ASC"
    UPDATED_AT_DESC = "UPDATED_AT_DESC"


class AddToFavoritesError(str, Enum):
    WEBAPP_NOT_FOUND = "WEBAPP_NOT_FOUND"


class ApproveAccessmodAccessRequestError(str, Enum):
    INVALID = "INVALID"


class ArchiveWorkspaceError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class BucketObjectType(str, Enum):
    DIRECTORY = "DIRECTORY"
    FILE = "FILE"


class ConnectionType(str, Enum):
    CUSTOM = "CUSTOM"
    DHIS2 = "DHIS2"
    GCS = "GCS"
    IASO = "IASO"
    POSTGRESQL = "POSTGRESQL"
    S3 = "S3"


class CreateAccessmodAccessibilityAnalysisError(str, Enum):
    NAME_DUPLICATE = "NAME_DUPLICATE"


class CreateAccessmodFileError(str, Enum):
    URI_DUPLICATE = "URI_DUPLICATE"


class CreateAccessmodFilesetError(str, Enum):
    NAME_DUPLICATE = "NAME_DUPLICATE"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class CreateAccessmodProjectError(str, Enum):
    NAME_DUPLICATE = "NAME_DUPLICATE"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class CreateAccessmodProjectMemberError(str, Enum):
    ALREADY_EXISTS = "ALREADY_EXISTS"
    NOT_FOUND = "NOT_FOUND"
    NOT_IMPLEMENTED = "NOT_IMPLEMENTED"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class CreateAccessmodZonalStatisticsError(str, Enum):
    NAME_DUPLICATE = "NAME_DUPLICATE"


class CreateBucketFolderError(str, Enum):
    ALREADY_EXISTS = "ALREADY_EXISTS"
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class CreateConnectionError(str, Enum):
    INVALID_SLUG = "INVALID_SLUG"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class CreateDatasetError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class CreateDatasetVersionError(str, Enum):
    DATASET_NOT_FOUND = "DATASET_NOT_FOUND"
    DUPLICATE_NAME = "DUPLICATE_NAME"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class CreateDatasetVersionFileError(str, Enum):
    ALREADY_EXISTS = "ALREADY_EXISTS"
    INVALID_URI = "INVALID_URI"
    LOCKED_VERSION = "LOCKED_VERSION"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    VERSION_NOT_FOUND = "VERSION_NOT_FOUND"


class CreateMembershipError(str, Enum):
    ALREADY_EXISTS = "ALREADY_EXISTS"
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class CreatePipelineFromTemplateVersionError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    PIPELINE_TEMPLATE_VERSION_NOT_FOUND = "PIPELINE_TEMPLATE_VERSION_NOT_FOUND"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class CreatePipelineTemplateVersionError(str, Enum):
    DUPLICATE_TEMPLATE_NAME_OR_CODE = "DUPLICATE_TEMPLATE_NAME_OR_CODE"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    PIPELINE_NOT_FOUND = "PIPELINE_NOT_FOUND"
    PIPELINE_VERSION_NOT_FOUND = "PIPELINE_VERSION_NOT_FOUND"
    UNKNOWN_ERROR = "UNKNOWN_ERROR"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class CreateTeamError(str, Enum):
    NAME_DUPLICATE = "NAME_DUPLICATE"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class CreateTemplateVersionPermissionReason(str, Enum):
    NO_NEW_TEMPLATE_VERSION_AVAILABLE = "NO_NEW_TEMPLATE_VERSION_AVAILABLE"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    PIPELINE_IS_ALREADY_FROM_TEMPLATE = "PIPELINE_IS_ALREADY_FROM_TEMPLATE"
    PIPELINE_IS_NOTEBOOK = "PIPELINE_IS_NOTEBOOK"


class CreateWebappError(str, Enum):
    ALREADY_EXISTS = "ALREADY_EXISTS"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class CreateWorkspaceError(str, Enum):
    INVALID_SLUG = "INVALID_SLUG"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DAGRunOrderBy(str, Enum):
    EXECUTION_DATE_ASC = "EXECUTION_DATE_ASC"
    EXECUTION_DATE_DESC = "EXECUTION_DATE_DESC"


class DAGRunStatus(str, Enum):
    failed = "failed"
    queued = "queued"
    running = "running"
    stopped = "stopped"
    success = "success"
    terminating = "terminating"


class DAGRunTrigger(str, Enum):
    MANUAL = "MANUAL"
    SCHEDULED = "SCHEDULED"


class DHIS2ConnectionError(str, Enum):
    REQUEST_ERROR = "REQUEST_ERROR"
    UNKNOWN_ERROR = "UNKNOWN_ERROR"


class DHIS2ConnectionStatus(str, Enum):
    DOWN = "DOWN"
    UNKNOWN = "UNKNOWN"
    UP = "UP"


class DHIS2MetadataType(str, Enum):
    DATASETS = "DATASETS"
    DATA_ELEMENTS = "DATA_ELEMENTS"
    DATA_ELEMENT_GROUPS = "DATA_ELEMENT_GROUPS"
    INDICATORS = "INDICATORS"
    INDICATOR_GROUPS = "INDICATOR_GROUPS"
    ORG_UNITS = "ORG_UNITS"
    ORG_UNIT_GROUPS = "ORG_UNIT_GROUPS"
    ORG_UNIT_LEVELS = "ORG_UNIT_LEVELS"


class DeclineWorkspaceInvitationError(str, Enum):
    INVITATION_NOT_FOUND = "INVITATION_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteAccessmodAnalysisError(str, Enum):
    DELETE_FAILED = "DELETE_FAILED"
    NOT_FOUND = "NOT_FOUND"


class DeleteAccessmodFilesetError(str, Enum):
    FILESET_IN_USE = "FILESET_IN_USE"
    NOT_FOUND = "NOT_FOUND"


class DeleteAccessmodProjectError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteAccessmodProjectMemberError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    NOT_IMPLEMENTED = "NOT_IMPLEMENTED"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteBucketObjectError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteConnectionError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteDatasetError(str, Enum):
    DATASET_NOT_FOUND = "DATASET_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteDatasetLinkError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteDatasetVersionError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    VERSION_NOT_FOUND = "VERSION_NOT_FOUND"


class DeleteMembershipError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteMetadataAttributeError(str, Enum):
    METADATA_ATTRIBUTE_NOT_FOUND = "METADATA_ATTRIBUTE_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    TARGET_NOT_FOUND = "TARGET_NOT_FOUND"


class DeletePipelineVersionError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    PIPELINE_NOT_FOUND = "PIPELINE_NOT_FOUND"
    PIPELINE_VERSION_NOT_FOUND = "PIPELINE_VERSION_NOT_FOUND"


class DeleteTeamError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteTemplateVersionError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    TEMPLATE_VERSION_NOT_FOUND = "TEMPLATE_VERSION_NOT_FOUND"


class DeleteWebappError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    WEBAPP_NOT_FOUND = "WEBAPP_NOT_FOUND"


class DeleteWorkspaceDatabaseTableError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    TABLE_NOT_FOUND = "TABLE_NOT_FOUND"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class DeleteWorkspaceError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteWorkspaceInvitationError(str, Enum):
    INVITATION_NOT_FOUND = "INVITATION_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DeleteWorkspaceMemberError(str, Enum):
    MEMBERSHIP_NOT_FOUND = "MEMBERSHIP_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class DenyAccessmodAccessRequestError(str, Enum):
    INVALID = "INVALID"


class DisableTwoFactorError(str, Enum):
    INVALID_OTP = "INVALID_OTP"
    NOT_ENABLED = "NOT_ENABLED"


class EnableTwoFactorError(str, Enum):
    ALREADY_ENABLED = "ALREADY_ENABLED"
    EMAIL_MISMATCH = "EMAIL_MISMATCH"


class FileSampleStatus(str, Enum):
    FAILED = "FAILED"
    FINISHED = "FINISHED"
    PROCESSING = "PROCESSING"


class FileType(str, Enum):
    directory = "directory"
    file = "file"


class GenerateChallengeError(str, Enum):
    CHALLENGE_ERROR = "CHALLENGE_ERROR"
    DEVICE_NOT_FOUND = "DEVICE_NOT_FOUND"


class GenerateNewDatabasePasswordError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class GeneratePipelineWebhookUrlError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    PIPELINE_NOT_FOUND = "PIPELINE_NOT_FOUND"
    WEBHOOK_NOT_ENABLED = "WEBHOOK_NOT_ENABLED"


class GenerateWorkspaceTokenError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class IASOConnectionError(str, Enum):
    REQUEST_ERROR = "REQUEST_ERROR"
    UNKNOWN_ERROR = "UNKNOWN_ERROR"


class IASOMetadataType(str, Enum):
    IASO_FORMS = "IASO_FORMS"
    IASO_ORG_UNITS = "IASO_ORG_UNITS"
    IASO_PROJECTS = "IASO_PROJECTS"


class InviteWorkspaceMembershipError(str, Enum):
    ALREADY_EXISTS = "ALREADY_EXISTS"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    USER_NOT_FOUND = "USER_NOT_FOUND"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class JoinWorkspaceError(str, Enum):
    ALREADY_ACCEPTED = "ALREADY_ACCEPTED"
    ALREADY_EXISTS = "ALREADY_EXISTS"
    INVITATION_NOT_FOUND = "INVITATION_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class LaunchAccessmodAnalysisError(str, Enum):
    LAUNCH_FAILED = "LAUNCH_FAILED"


class LaunchNotebookServerError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class LinkDatasetError(str, Enum):
    ALREADY_LINKED = "ALREADY_LINKED"
    DATASET_NOT_FOUND = "DATASET_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class LoginError(str, Enum):
    INVALID_CREDENTIALS = "INVALID_CREDENTIALS"
    INVALID_OTP = "INVALID_OTP"
    OTP_REQUIRED = "OTP_REQUIRED"


class MembershipRole(str, Enum):
    ADMIN = "ADMIN"
    REGULAR = "REGULAR"


class MessagePriority(str, Enum):
    CRITICAL = "CRITICAL"
    DEBUG = "DEBUG"
    ERROR = "ERROR"
    INFO = "INFO"
    WARNING = "WARNING"


class OrderByDirection(str, Enum):
    ASC = "ASC"
    DESC = "DESC"


class ParameterType(str, Enum):
    bool = "bool"
    custom = "custom"
    dataset = "dataset"
    dhis2 = "dhis2"
    float = "float"
    gcs = "gcs"
    iaso = "iaso"
    int = "int"
    postgresql = "postgresql"
    s3 = "s3"
    str = "str"


class ParameterWidget(str, Enum):
    DHIS2_DATASETS = "DHIS2_DATASETS"
    DHIS2_DATA_ELEMENTS = "DHIS2_DATA_ELEMENTS"
    DHIS2_DATA_ELEMENT_GROUPS = "DHIS2_DATA_ELEMENT_GROUPS"
    DHIS2_INDICATORS = "DHIS2_INDICATORS"
    DHIS2_INDICATOR_GROUPS = "DHIS2_INDICATOR_GROUPS"
    DHIS2_ORG_UNITS = "DHIS2_ORG_UNITS"
    DHIS2_ORG_UNIT_GROUPS = "DHIS2_ORG_UNIT_GROUPS"
    DHIS2_ORG_UNIT_LEVELS = "DHIS2_ORG_UNIT_LEVELS"
    IASO_FORMS = "IASO_FORMS"
    IASO_ORG_UNITS = "IASO_ORG_UNITS"
    IASO_PROJECTS = "IASO_PROJECTS"


class PermissionMode(str, Enum):
    EDITOR = "EDITOR"
    OWNER = "OWNER"
    VIEWER = "VIEWER"


class PinDatasetError(str, Enum):
    LINK_NOT_FOUND = "LINK_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class PipelineError(str, Enum):
    CANNOT_UPDATE_NOTEBOOK_PIPELINE = "CANNOT_UPDATE_NOTEBOOK_PIPELINE"
    DUPLICATE_PIPELINE_VERSION_NAME = "DUPLICATE_PIPELINE_VERSION_NAME"
    FILE_NOT_FOUND = "FILE_NOT_FOUND"
    INVALID_CONFIG = "INVALID_CONFIG"
    INVALID_TIMEOUT_VALUE = "INVALID_TIMEOUT_VALUE"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    PIPELINE_ALREADY_COMPLETED = "PIPELINE_ALREADY_COMPLETED"
    PIPELINE_ALREADY_STOPPED = "PIPELINE_ALREADY_STOPPED"
    PIPELINE_DOES_NOT_SUPPORT_PARAMETERS = "PIPELINE_DOES_NOT_SUPPORT_PARAMETERS"
    PIPELINE_NOT_FOUND = "PIPELINE_NOT_FOUND"
    PIPELINE_VERSION_NOT_FOUND = "PIPELINE_VERSION_NOT_FOUND"
    TABLE_NOT_FOUND = "TABLE_NOT_FOUND"
    WORKSPACE_NOT_FOUND = "WORKSPACE_NOT_FOUND"


class PipelineNotificationLevel(str, Enum):
    ALL = "ALL"
    ERROR = "ERROR"


class PipelineRecipientError(str, Enum):
    ALREADY_EXISTS = "ALREADY_EXISTS"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    PIPELINE_NOT_FOUND = "PIPELINE_NOT_FOUND"
    RECIPIENT_NOT_FOUND = "RECIPIENT_NOT_FOUND"
    USER_NOT_FOUND = "USER_NOT_FOUND"


class PipelineRunOrderBy(str, Enum):
    EXECUTION_DATE_ASC = "EXECUTION_DATE_ASC"
    EXECUTION_DATE_DESC = "EXECUTION_DATE_DESC"


class PipelineRunStatus(str, Enum):
    failed = "failed"
    queued = "queued"
    running = "running"
    stopped = "stopped"
    success = "success"
    terminating = "terminating"


class PipelineRunTrigger(str, Enum):
    manual = "manual"
    scheduled = "scheduled"
    webhook = "webhook"


class PipelineTemplateError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    PIPELINE_TEMPLATE_NOT_FOUND = "PIPELINE_TEMPLATE_NOT_FOUND"


class PipelineType(str, Enum):
    notebook = "notebook"
    zipFile = "zipFile"


class PrepareObjectDownloadError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class PrepareObjectUploadError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"


class PrepareVersionFileDownloadError(str, Enum):
    FILE_NOT_FOUND = "FILE_NOT_FOUND"
    FILE_NOT_UPLOADED = "FILE_NOT_UPLOADED"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class RegisterError(str, Enum):
    ALREADY_LOGGED_IN = "ALREADY_LOGGED_IN"
    EMAIL_TAKEN = "EMAIL_TAKEN"
    INVALID_PASSWORD = "INVALID_PASSWORD"
    INVALID_TOKEN = "INVALID_TOKEN"
    PASSWORD_MISMATCH = "PASSWORD_MISMATCH"


class RemoveFromFavoritesError(str, Enum):
    WEBAPP_NOT_FOUND = "WEBAPP_NOT_FOUND"


class RequestAccessmodAccessError(str, Enum):
    ALREADY_EXISTS = "ALREADY_EXISTS"
    INVALID = "INVALID"
    MUST_ACCEPT_TOS = "MUST_ACCEPT_TOS"


class ResendWorkspaceInvitationError(str, Enum):
    INVITATION_NOT_FOUND = "INVITATION_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class RunDAGError(str, Enum):
    DAG_NOT_FOUND = "DAG_NOT_FOUND"
    INVALID_CONFIG = "INVALID_CONFIG"


class SetDAGRunFavoriteError(str, Enum):
    INVALID = "INVALID"
    MISSING_LABEL = "MISSING_LABEL"
    NOT_FOUND = "NOT_FOUND"


class SetMetadataAttributeError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    TARGET_NOT_FOUND = "TARGET_NOT_FOUND"


class SetPasswordError(str, Enum):
    INVALID_PASSWORD = "INVALID_PASSWORD"
    INVALID_TOKEN = "INVALID_TOKEN"
    PASSWORD_MISMATCH = "PASSWORD_MISMATCH"
    USER_NOT_FOUND = "USER_NOT_FOUND"


class UpdateAccessmodAccessibilityAnalysisError(str, Enum):
    NAME_DUPLICATE = "NAME_DUPLICATE"
    NOT_FOUND = "NOT_FOUND"


class UpdateAccessmodFilesetError(str, Enum):
    NAME_DUPLICATE = "NAME_DUPLICATE"
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateAccessmodProjectError(str, Enum):
    NAME_DUPLICATE = "NAME_DUPLICATE"
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateAccessmodProjectMemberError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    NOT_IMPLEMENTED = "NOT_IMPLEMENTED"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateAccessmodZonalStatisticsError(str, Enum):
    NAME_DUPLICATE = "NAME_DUPLICATE"
    NOT_FOUND = "NOT_FOUND"


class UpdateConnectionError(str, Enum):
    INVALID_SLUG = "INVALID_SLUG"
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateDAGError(str, Enum):
    INVALID = "INVALID"
    NOT_FOUND = "NOT_FOUND"


class UpdateDatasetError(str, Enum):
    DATASET_NOT_FOUND = "DATASET_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateDatasetVersionError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    VERSION_NOT_FOUND = "VERSION_NOT_FOUND"


class UpdateMembershipError(str, Enum):
    INVALID_ROLE = "INVALID_ROLE"
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdatePipelineError(str, Enum):
    INVALID_CONFIG = "INVALID_CONFIG"
    MISSING_VERSION_CONFIG = "MISSING_VERSION_CONFIG"
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdatePipelineVersionError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateTeamError(str, Enum):
    NAME_DUPLICATE = "NAME_DUPLICATE"
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateTemplateError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateTemplateVersionError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateUserError(str, Enum):
    INVALID_LANGUAGE = "INVALID_LANGUAGE"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateWebappError(str, Enum):
    PERMISSION_DENIED = "PERMISSION_DENIED"
    WEBAPP_NOT_FOUND = "WEBAPP_NOT_FOUND"


class UpdateWorkspaceError(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpdateWorkspaceMemberError(str, Enum):
    MEMBERSHIP_NOT_FOUND = "MEMBERSHIP_NOT_FOUND"
    PERMISSION_DENIED = "PERMISSION_DENIED"


class UpgradePipelineVersionFromTemplateError(str, Enum):
    NO_NEW_TEMPLATE_VERSION_AVAILABLE = "NO_NEW_TEMPLATE_VERSION_AVAILABLE"
    PIPELINE_NOT_FOUND = "PIPELINE_NOT_FOUND"
    PIPELINE_NOT_FROM_TEMPLATE = "PIPELINE_NOT_FROM_TEMPLATE"


class VerifyDeviceError(str, Enum):
    INVALID_OTP = "INVALID_OTP"
    NO_DEVICE = "NO_DEVICE"


class WorkspaceInvitationStatus(str, Enum):
    ACCEPTED = "ACCEPTED"
    DECLINED = "DECLINED"
    PENDING = "PENDING"


class WorkspaceMembershipRole(str, Enum):
    ADMIN = "ADMIN"
    EDITOR = "EDITOR"
    VIEWER = "VIEWER"



# Generated by ariadne-codegen
# Source: openhexa/cli/graphql/queries.graphql

from pydantic import Field

from .base_model import BaseModel


class Me(BaseModel):
    me: "MeMe"


class MeMe(BaseModel):
    has_two_factor_enabled: bool = Field(alias="hasTwoFactorEnabled")


Me.model_rebuild()



"""Example pipeline to get data elements from IASO."""

from openhexa.sdk import current_run, parameter, pipeline
from openhexa.sdk.pipelines.parameter import IASOWidget
from openhexa.sdk.workspaces.connection import IASOConnection


@pipeline(name="IASO Data Elements")
@parameter("iaso_con", type=IASOConnection, required=True)
@parameter(
    "forms",
    type=int,
    required=True,
    widget=IASOWidget.IASO_FORMS,
    multiple=True,
    connection="iaso_con",
)
@parameter(
    "projects",
    type=int,
    required=True,
    widget=IASOWidget.IASO_PROJECTS,
    multiple=True,
    connection="iaso_con",
)
@parameter(
    "org_units",
    type=int,
    required=True,
    widget=IASOWidget.IASO_ORG_UNITS,
    multiple=True,
    connection="iaso_con",
)
def iaso(iaso_con, forms, org_units, projects):
    """Get forms from IASO."""
    print_forms(iaso_con, forms, org_units, projects)


@iaso.task
def print_forms(iaso_con, forms, org_units, projects):
    """Print forms."""
    current_run.log_info("Printing forms")

    current_run.log_info(f"Forms: {forms}, Org Units: {org_units}, Projects: {projects}")


if __name__ == "__main__":
    iaso()



"""CLI test module."""

import base64
import os
from io import BytesIO
from pathlib import Path
from tempfile import mkdtemp
from unittest import TestCase
from unittest.mock import MagicMock, patch
from zipfile import ZipFile

import pytest
from click.testing import CliRunner

from openhexa.cli.cli import pipelines_download, pipelines_push, pipelines_run, select_pipeline, workspaces_add
from openhexa.sdk.pipelines import Pipeline

python_file_name = "pipeline.py"
python_code = "print('pipeline.py file')"
version = "v1.0"
pipeline_name = "MyPipeline"
pipeline_id = "pipeline_id"
pipeline_version_id = "pipeline_version_id"
template = {
    "id": "template_id",
    "name": "test_template",
    "code": "template_code",
    "currentVersion": {"versionNumber": "1.0"},
}
changelog = "Changelog for the new version"


def create_zip_with_pipeline():
    """Create a zip file containing the pipeline.py file."""
    zip_buffer = BytesIO()
    fake_zipfile = ZipFile(zip_buffer, "w")
    fake_zipfile.writestr(python_file_name, python_code)
    fake_zipfile.close()
    return zip_buffer


def setup_graphql_response(zip_buffer=create_zip_with_pipeline()):
    """Set up the mock GraphQL response pipelines."""
    return {
        "pipelineByCode": {
            "code": "pipeline-1234",
            "currentVersion": {"zipfile": base64.b64encode(zip_buffer.getvalue()).decode()},
        },
        "pipelines": {"items": []},  # (empty workspace initially)
    }


@pytest.mark.usefixtures("settings")
class CliRunTest(TestCase):
    """Test the CLI."""

    runner = None

    def setUp(self):
        """Configure the CLI runner and the user config."""
        self.runner = CliRunner()
        return super().setUp()

    def tearDown(self) -> None:
        """Tear down the CLI runner and the user config.

        Returns
        -------
            None
        """
        return super().tearDown()

    @patch("openhexa.cli.api.ask_pipeline_config_creation", return_value=True)
    def test_no_pipeline(self, *args):
        """Test running a pipeline without a pipeline.py file."""
        with self.runner.isolated_filesystem():
            result = self.runner.invoke(pipelines_run, [mkdtemp()])
            assert result.exit_code == 1
            self.assertTrue("does not contain a pipeline.py file" in str(result.exception))

    @patch("openhexa.cli.api.graphql")
    def test_download_pipeline_no_pipeline(self, mock_graphql):
        """Test the download pipeline command."""
        with self.runner.isolated_filesystem() as tmp:
            # Unknown pipeline
            mock_graphql.return_value = {"pipelineByCode": None}
            result = self.runner.invoke(pipelines_download, ["test_pipeline", tmp])
            self.assertIn(
                "No pipeline exists in workspace-slug with code test_pipeline",
                str(result.exception),
            )

    @patch("openhexa.cli.api.graphql")
    def test_download_pipeline_overwrite(self, mock_graphql):
        """Test the download pipeline command with an existing director with and without content."""
        with self.runner.isolated_filesystem() as tmp:
            # Known Pipeline & non-empty directory
            with open(Path(tmp) / python_file_name, "w") as f:
                f.write("<content>")

            mock_graphql.return_value = setup_graphql_response()

            result = self.runner.invoke(pipelines_download, ["test_pipeline", tmp], input="N\n")
            self.assertIn("Overwrite the files?", result.output)
            self.assertIn(f"{tmp} is not empty", result.output)
            self.assertEqual(open(Path(tmp) / python_file_name).read(), "<content>")

            # Overwrite the files in the directory
            mock_graphql.return_value = setup_graphql_response()

            result = self.runner.invoke(pipelines_download, ["test_pipeline", tmp], input="y\n")
            self.assertIn("Overwrite the files?", result.output)
            self.assertEqual(open(Path(tmp) / python_file_name).read(), python_code)

    @patch("openhexa.cli.api.graphql")
    def test_download_pipeline(self, mock_graphql):
        """Test the download pipeline command."""
        with self.runner.isolated_filesystem() as tmp:
            # Create a zipfile with a pipeline.py file in a buffer
            mock_graphql.return_value = setup_graphql_response()

            result = self.runner.invoke(pipelines_download, ["test_pipeline", tmp])
            self.assertEqual(result.exit_code, 0)
            path = Path(tmp) / python_file_name
            self.assertTrue(path.exists())
            self.assertEqual(open(path).read(), python_code)

    @patch("openhexa.cli.api.graphql")
    @patch("openhexa.cli.cli.get_pipeline")
    @patch("openhexa.cli.cli.get_pipelines_pages")
    @patch("openhexa.cli.cli.upload_pipeline")
    @patch("openhexa.cli.cli.create_pipeline_template_version")
    @patch.dict(os.environ, {"HEXA_API_URL": "https://www.bluesquarehub.com/", "HEXA_WORKSPACE": "workspace"})
    def test_push_pipeline(
        self, mock_create_template, mock_upload_pipeline, mock_get_pipelines_pages, mock_get_pipeline, mock_graphql
    ):
        """Test pushing a pipeline."""
        with self.runner.isolated_filesystem() as tmp:
            with open(Path(tmp) / python_file_name, "w") as f:
                f.write(python_code)
            mock_graphql.return_value = setup_graphql_response()
            mock_pipeline = MagicMock(spec=Pipeline)
            mock_pipeline.name = pipeline_name
            mock_get_pipeline.return_value = mock_pipeline
            mock_get_pipelines_pages.return_value = {
                "items": [
                    {"name": "Pipeline1", "code": "code1"},
                    {"name": "Pipeline2", "code": "code2"},
                ],
                "totalPages": 2,
            }
            mock_upload_pipeline.return_value = {
                "versionName": version,
                "pipeline": {
                    "id": pipeline_id,
                    "permissions": {"createTemplateVersion": {"isAllowed": True}},
                    "template": template,
                },
                "id": pipeline_version_id,
            }
            mock_create_template.return_value = template

            result = self.runner.invoke(
                pipelines_push,
                [tmp, "--name", version],
                input="\n".join(["4", "pipeline_code", "Y", "Y", changelog]) + "\n",
            )
            self.assertEqual(result.exit_code, 0)
            self.assertIn("Which pipeline do you want to update?", result.output)
            self.assertIn("Insert a pipeline code", result.output)
            self.assertIn(
                (
                    f"✅ New version '{version}' created! "
                    f"You can view the pipeline in OpenHEXA on https://www.bluesquarehub.com/workspaces/workspace/pipelines/pipeline-1234"
                ),
                result.output,
            )
            self.assertTrue(mock_upload_pipeline.called)
            mock_create_template.assert_called_with("workspace", pipeline_id, pipeline_version_id, changelog)
            self.assertIn(
                (
                    f"✅ New version '{template['currentVersion']['versionNumber']}' of the template '{template['name']}' created! "
                    f"You can view the new template version in OpenHEXA on https://www.bluesquarehub.com/workspaces/workspace/templates/{template['code']}/versions"
                ),
                result.output,
            )

    @patch("openhexa.cli.api.graphql")
    @patch.dict(os.environ, {"HEXA_API_URL": "https://www.bluesquarehub.com/", "HEXA_WORKSPACE": "workspace"})
    def test_push_pipeline_with_yes_flag_without_code(self, mock_graphql):
        """Test pushing a pipeline with the --yes flag without providing a --code."""
        with self.runner.isolated_filesystem() as tmp:
            with open(Path(tmp) / python_file_name, "w") as f:
                f.write(python_code)
            mock_graphql.return_value = setup_graphql_response()

            result = self.runner.invoke(
                pipelines_push,
                [tmp, "--yes"],
            )
            self.assertEqual(result.exit_code, 1)
            self.assertIn(
                "❌ You must provide a pipeline code (using -c or --code) when using the --yes flag.", result.output
            )

    @patch("openhexa.cli.api.graphql")
    @patch("openhexa.cli.cli.get_pipeline")
    @patch("openhexa.cli.cli.get_pipelines_pages")
    @patch("openhexa.cli.cli.get_pipeline_from_code")
    @patch.dict(os.environ, {"HEXA_API_URL": "https://www.bluesquarehub.com/", "HEXA_WORKSPACE": "workspace"})
    def test_push_pipeline_with_non_existing_code(
        self, mock_get_pipeline_from_code, mock_get_pipelines_pages, mock_get_pipeline, mock_graphql
    ):
        """Test pushing a pipeline with a non-existing --code flag."""
        with self.runner.isolated_filesystem() as tmp:
            with open(Path(tmp) / python_file_name, "w") as f:
                f.write(python_code)
            mock_graphql.return_value = setup_graphql_response()
            mock_pipeline = MagicMock(spec=Pipeline)
            mock_pipeline.name = pipeline_name
            mock_get_pipeline.return_value = mock_pipeline
            mock_get_pipelines_pages.return_value = {
                "items": [
                    {"name": "Pipeline1", "code": "code1"},
                    {"name": "Pipeline2", "code": "code2"},
                ],
                "totalPages": 2,
            }
            mock_get_pipeline_from_code.return_value = None  # Simulate non-existing pipeline code

            result = self.runner.invoke(
                pipelines_push,
                [tmp, "--code", "non_existing_code"],
            )
            self.assertEqual(result.exit_code, 1)
            self.assertIn("❌ Pipeline with code 'non_existing_code' not found.", result.output)

    @patch("openhexa.cli.api.graphql")
    @patch("openhexa.cli.cli.get_pipeline")
    @patch("openhexa.cli.cli.get_pipelines_pages")
    @patch("openhexa.cli.cli.upload_pipeline")
    @patch.dict(os.environ, {"HEXA_API_URL": "https://www.bluesquarehub.com/", "HEXA_WORKSPACE": "workspace"})
    def test_push_pipeline_with_code_flag(
        self, mock_upload_pipeline, mock_get_pipelines_pages, mock_get_pipeline, mock_graphql
    ):
        """Test pushing a pipeline using the --code flag."""
        code = "code1"
        with self.runner.isolated_filesystem() as tmp:
            with open(Path(tmp) / python_file_name, "w") as f:
                f.write(python_code)
            mock_graphql.return_value = {
                "pipelineByCode": {
                    "code": code,
                    "currentVersion": {"zipfile": ""},
                },
                "pipelines": {"items": []},
            }
            mock_pipeline = MagicMock(spec=Pipeline)
            mock_pipeline.name = pipeline_name
            mock_get_pipeline.return_value = mock_pipeline
            mock_get_pipelines_pages.return_value = {
                "items": [
                    {"name": "Pipeline1", "code": "code1"},
                    {"name": "Pipeline2", "code": "code2"},
                ],
                "totalPages": 2,
            }
            mock_upload_pipeline.return_value = {
                "versionName": version,
                "pipeline": {
                    "id": pipeline_id,
                    "permissions": {"createTemplateVersion": {"isAllowed": True}},
                    "template": template,
                },
                "id": pipeline_version_id,
            }

            result = self.runner.invoke(
                pipelines_push,
                [tmp, "--name", version, "--code", code],
                input="Y\nn\n",
            )
            self.assertEqual(result.exit_code, 0)
            self.assertNotIn("Which pipeline do you want to update?", result.output)
            self.assertTrue(mock_upload_pipeline.called)
            self.assertEqual(mock_upload_pipeline.call_args[0][0], code)
            self.assertIn(
                f"✅ New version '{version}' created! ",
                result.output,
            )

    @patch("openhexa.cli.cli.click.prompt")
    def test_select_pipeline(self, mock_prompt):
        workspace_pipelines = [
            {"name": "Pipeline1", "code": "code1"},
            {"name": "Pipeline2", "code": "code2"},
        ]
        pipeline = MagicMock()
        pipeline.name = "TestPipeline"

        mock_prompt.side_effect = [1]  # User selects the first pipeline

        selected_pipeline = select_pipeline(workspace_pipelines, 1, pipeline)

        self.assertEqual(selected_pipeline, workspace_pipelines[0])

    @patch("openhexa.cli.cli.click.prompt")
    def test_select_pipeline_create_new(self, mock_prompt):
        workspace_pipelines = []
        pipeline = MagicMock()
        pipeline.name = "TestPipeline"

        mock_prompt.side_effect = [1]  # User selects to create a new pipeline

        selected_pipeline = select_pipeline(workspace_pipelines, 1, pipeline)

        self.assertIsNone(selected_pipeline)

    @patch("openhexa.cli.cli.get_pipeline_from_code")
    @patch("openhexa.cli.cli.click.prompt")
    def test_select_pipeline_enter_code(self, mock_prompt, mock_get_pipeline_from_code):
        workspace_pipelines = []
        pipeline = MagicMock()
        pipeline.name = "TestPipeline"
        pipeline = MagicMock()
        pipeline.name = "Pipeline3"
        pipeline.code = "code3"
        mock_prompt.side_effect = [2, pipeline.code]  # User selects to enter a pipeline code and provides "code3"
        mock_get_pipeline_from_code.return_value = pipeline

        selected_pipeline = select_pipeline(workspace_pipelines, 2, pipeline)

        self.assertEqual(selected_pipeline, pipeline)
        mock_get_pipeline_from_code.assert_called_with(pipeline.code)

    @patch("openhexa.cli.api.graphql")
    def test_workspaces_add_not_found(self, mock_graphql):
        """Test the add workspace command when the workspace doesn't exist on the current server."""
        with self.runner.isolated_filesystem():
            mock_graphql.return_value = {"workspace": None}
            result = self.runner.invoke(workspaces_add, ["test_workspace"], input="random_token \n")
            self.assertEqual(result.exit_code, 1)
            self.assertIn(
                "Workspace test_workspace does not exist",
                str(result.stdout),
            )

    @patch("openhexa.cli.api.graphql")
    def test_workspaces_add(self, mock_graphql):
        """Test the add workspace command."""
        with self.runner.isolated_filesystem():
            mock_graphql.return_value = {"workspace": {"name": "test_workspace", "slug": "test_workspace_0000"}}
            result = self.runner.invoke(workspaces_add, ["test_workspace"], input="random_token \n")
            self.assertEqual(result.exit_code, 0)



"""Pipeline test module."""

import os
from unittest import TestCase
from unittest.mock import Mock, patch

import pytest

from openhexa.sdk import (
    DHIS2Connection,
    GCSConnection,
    IASOConnection,
    PostgreSQLConnection,
    S3Connection,
)
from openhexa.sdk.pipelines.log_level import LogLevel
from openhexa.sdk.pipelines.parameter import Parameter, ParameterValueError
from openhexa.sdk.pipelines.pipeline import Pipeline


def test_pipeline_run_valid_config():
    """Happy path for pipeline run config."""
    pipeline_func = Mock()
    parameter_1 = Parameter("arg1", type=str)
    parameter_2 = Parameter("arg2", type=str, multiple=True)
    parameter_3 = Parameter("arg3", type=int, default=33)
    pipeline = Pipeline("pipeline", pipeline_func, [parameter_1, parameter_2, parameter_3])
    pipeline.run({"arg1": "ab", "arg2": ["cd", "ef"]})

    assert pipeline.name == "pipeline"
    pipeline_func.assert_called_once_with(arg1="ab", arg2=["cd", "ef"], arg3=33)


def test_pipeline_run_invalid_config():
    """Verify thatinvalid configuration values raise an exception."""
    pipeline_func = Mock()
    parameter_1 = Parameter("arg1", type=str)
    pipeline = Pipeline("pipeline", pipeline_func, [parameter_1])
    with pytest.raises(ParameterValueError):
        pipeline.run({"arg1": 3})


def test_pipeline_run_extra_config():
    """Verify that extra (unexpected) configuration values raise an exception."""
    pipeline_func = Mock()
    parameter_1 = Parameter("arg1", type=str)
    pipeline = Pipeline("pipeline", pipeline_func, [parameter_1])
    with pytest.raises(ParameterValueError):
        pipeline.run({"arg1": "ok", "arg2": "extra"})


@patch.dict(
    os.environ,
    {"HEXA_SERVER_URL": "https://test.openhexa.org"},
)
def test_pipeline_run_connection_dhis2_parameter_config(workspace):
    """Ensure that DHIS2 connection parameter values are built properly."""
    identifier = "dhis2-connection-id"

    url = "https://test.dhis2.org/"
    username = "dhis2"
    password = "dhis2_pwd"

    pipeline_func = Mock()
    parameter_1 = Parameter(
        "connection_param",
        name="this is a test for connection parameter",
        type=DHIS2Connection,
    )
    pipeline = Pipeline("pipeline", pipeline_func, [parameter_1])

    data = DHIS2Connection(url, username, password)
    with patch.object(workspace, "get_connection", return_value=data):
        pipeline.run({"connection_param": identifier})
        assert pipeline.name == "pipeline"
        pipeline_func.assert_called_once_with(connection_param=DHIS2Connection(url, username, password))


@patch.dict(
    os.environ,
    {"HEXA_SERVER_URL": "https://test.openhexa.org"},
)
def test_pipeline_run_connection_iaso_parameter_config(workspace):
    """Ensure that IASO connection parameter values are built properly."""
    identifier = "iaso-connection-id"

    url = "https://test.iaso.org/"
    username = "iaso"
    password = "iaso_pwd"

    pipeline_func = Mock()
    parameter_1 = Parameter(
        "connection_param",
        name="this is a test for connection parameter",
        type=IASOConnection,
    )
    pipeline = Pipeline("pipeline", pipeline_func, [parameter_1])

    data = IASOConnection(url, username, password)
    with patch.object(workspace, "get_connection", return_value=data):
        pipeline.run({"connection_param": identifier})
        assert pipeline.name == "pipeline"
        pipeline_func.assert_called_once_with(connection_param=IASOConnection(url, username, password))


@patch.dict(
    os.environ,
    {"HEXA_SERVER_URL": "https://test.openhexa.org"},
)
def test_pipeline_run_connection_gcs_parameter_config(workspace):
    """Ensure that GCS connection parameter values are built properly."""
    identifier = "gcs-connection-id"

    service_account_key = "HqQBxH0BAI3zF7kANUNlGg"
    bucket_name = "test"

    pipeline_func = Mock()
    parameter_1 = Parameter(
        "connection_param",
        name="this is a test for connection parameter",
        type=GCSConnection,
    )
    pipeline = Pipeline("pipeline", pipeline_func, [parameter_1])

    data = GCSConnection(service_account_key, bucket_name)

    with patch.object(workspace, "get_connection", return_value=data):
        pipeline.run({"connection_param": identifier})
        assert pipeline.name == "pipeline"
        pipeline_func.assert_called_once_with(connection_param=GCSConnection(service_account_key, bucket_name))


@patch.dict(
    os.environ,
    {"HEXA_SERVER_URL": "https://test.openhexa.org"},
)
def test_pipeline_run_connection_s3_parameter_config(workspace):
    """Ensure that S3 connection parameter values are built properly."""
    identifier = "s3-connection-id"

    secret_access_key = "HqQBxH0BAI3zF7kANUNlGg"
    access_key_id = "84hVntMaMSYP/RSW9ex04w"
    bucket_name = "test"

    pipeline_func = Mock()
    parameter_1 = Parameter(
        "connection_param",
        name="this is a test for connection parameter",
        type=S3Connection,
    )
    pipeline = Pipeline("pipeline", pipeline_func, [parameter_1])

    data = S3Connection(access_key_id, secret_access_key, bucket_name)
    with patch.object(workspace, "get_connection", return_value=data):
        pipeline.run({"connection_param": identifier})
        assert pipeline.name == "pipeline"
        pipeline_func.assert_called_once_with(
            connection_param=S3Connection(access_key_id, secret_access_key, bucket_name)
        )


@patch.dict(
    os.environ,
    {"HEXA_SERVER_URL": "https://test.openhexa.org"},
)
def test_pipeline_run_connection_postgres_parameter_config(workspace):
    """Ensure that postgreSQL connection parameter values are built properly."""
    identifier = "postgres-connection-id"

    host = "https://127.0.0.1"
    port = 5432
    username = "hexa_sdk"
    password = "hexa_sdk_pwd"
    database_name = "hexa_sdk"

    pipeline_func = Mock()
    parameter_1 = Parameter(
        "connection_param",
        name="this is a test for connection parameter",
        type=PostgreSQLConnection,
    )

    pipeline = Pipeline("pipeline", pipeline_func, [parameter_1])

    data = PostgreSQLConnection(host, port, username, password, database_name)
    with patch.object(workspace, "get_connection", return_value=data):
        pipeline.run({"connection_param": identifier})
        assert pipeline.name == "pipeline"
        pipeline_func.assert_called_once_with(
            connection_param=PostgreSQLConnection(host, port, username, password, database_name)
        )


class TestLogLevel(TestCase):
    def test_parse_log_level(self):
        test_cases = [
            (0, LogLevel.DEBUG),
            (1, LogLevel.INFO),
            (2, LogLevel.WARNING),
            (3, LogLevel.ERROR),
            (4, LogLevel.CRITICAL),
            ("0", LogLevel.DEBUG),
            ("1", LogLevel.INFO),
            ("2", LogLevel.WARNING),
            ("3", LogLevel.ERROR),
            ("4", LogLevel.CRITICAL),
            ("DEBUG", LogLevel.DEBUG),
            ("INFO", LogLevel.INFO),
            ("WARNING", LogLevel.WARNING),
            ("ERROR", LogLevel.ERROR),
            ("CRITICAL", LogLevel.CRITICAL),
            ("invalid", LogLevel.INFO),
            (6, LogLevel.INFO),
            (-1, LogLevel.INFO),
        ]
        for input_value, expected in test_cases:
            with self.subTest(input_value=input_value, expected=expected):
                self.assertEqual(LogLevel.parse_log_level(input_value), expected)



"""Tests related to the parsing of the pipeline code."""

import io
import tempfile
from unittest import TestCase
from unittest.mock import patch

from openhexa.sdk.pipelines.exceptions import InvalidParameterError, PipelineNotFound
from openhexa.sdk.pipelines.parameter import DHIS2Widget, IASOWidget
from openhexa.sdk.pipelines.runtime import get_pipeline


class AstTest(TestCase):
    """Test the parsing of the pipeline code."""

    def test_pipeline_not_found(self):
        """The file does not contain a @pipeline decorator."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write("print('hello')")

            with self.assertRaises(PipelineNotFound):
                get_pipeline(tmpdirname)

    def test_pipeline_no_parameters(self):
        """The file contains a @pipeline decorator but no parameters."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline",
                            "",
                            "@pipeline(name='Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "name": "Test pipeline",
                    "function": None,
                    "tasks": [],
                    "parameters": [],
                    "timeout": None,
                },
            )

    def test_pipeline_with_args(self):
        """The file contains a @pipeline decorator with args."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline",
                            "",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "function": None,
                    "tasks": [],
                    "name": "Test pipeline",
                    "parameters": [],
                    "timeout": None,
                },
            )

    def test_pipeline_with_invalid_parameter_args(self):
        """The file contains a @parameter decorator with a BinOp as default value."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "",
                            "@pipeline('Test pipeline')",
                            "@parameter('test_param', name='Test Param', type=int, default=42 * 30, help='Param help')",
                            "def test_pipeline(test_param):",
                            "    pass",
                        ]
                    )
                )
            with self.assertRaises(ValueError):
                get_pipeline(tmpdirname)

    def test_pipeline_with_invalid_pipeline_args(self):
        """The file contains a @pipeline decorator with invalid value."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "",
                            "timeout = 60 * 60",
                            "@pipeline('Test pipeline', timeout=timeout)",
                            "@parameter('test_param', name='Test Param', type=int, help='Param help')",
                            "def test_pipeline(test_param):",
                            "    pass",
                        ]
                    )
                )
            with self.assertRaises(ValueError):
                get_pipeline(tmpdirname)

    def test_pipeline_with_int_param(self):
        """The file contains a @pipeline decorator and a @parameter decorator with an int."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "",
                            "@parameter('test_param', name='Test Param', type=int, default=42, help='Param help')",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "name": "Test pipeline",
                    "function": None,
                    "tasks": [],
                    "parameters": [
                        {
                            "choices": None,
                            "multiple": False,
                            "code": "test_param",
                            "type": "int",
                            "name": "Test Param",
                            "default": 42,
                            "widget": None,
                            "connection": None,
                            "help": "Param help",
                            "required": True,
                        }
                    ],
                    "timeout": None,
                },
            )

    def test_pipeline_with_multiple_param(self):
        """The file contains a @pipeline decorator and a @parameter decorator with multiple=True."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "",
                            "@parameter('test_param', name='Test Param', type=int, default=[42], help='Param help', multiple=True)",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "name": "Test pipeline",
                    "function": None,
                    "tasks": [],
                    "parameters": [
                        {
                            "choices": None,
                            "multiple": True,
                            "code": "test_param",
                            "type": "int",
                            "name": "Test Param",
                            "default": [42],
                            "widget": None,
                            "connection": None,
                            "help": "Param help",
                            "required": True,
                        }
                    ],
                    "timeout": None,
                },
            )

    def test_pipeline_with_dataset(self):
        """The file contains a @pipeline decorator and use a parameter with 'dataset' type."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "from openhexa.sdk.pipelines.parameter import Dataset",
                            "",
                            "@parameter('dataset', name='Dataset', type=Dataset, help='Dataset', required=False)",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "function": None,
                    "name": "Test pipeline",
                    "tasks": [],
                    "parameters": [
                        {
                            "choices": None,
                            "multiple": False,
                            "code": "dataset",
                            "type": "dataset",
                            "name": "Dataset",
                            "default": None,
                            "widget": None,
                            "connection": None,
                            "help": "Dataset",
                            "required": False,
                        }
                    ],
                    "timeout": None,
                },
            )

    def test_pipeline_with_choices(self):
        """The file contains a @pipeline decorator and a @parameter decorator with choices."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "",
                            "@parameter('test_param', name='Test Param', type=str, choices=['a', 'b'], help='Param help')",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "name": "Test pipeline",
                    "function": None,
                    "tasks": [],
                    "parameters": [
                        {
                            "choices": ["a", "b"],
                            "multiple": False,
                            "code": "test_param",
                            "type": "str",
                            "name": "Test Param",
                            "default": None,
                            "widget": None,
                            "connection": None,
                            "help": "Param help",
                            "required": True,
                        }
                    ],
                    "timeout": None,
                },
            )

    def test_pipeline_with_timeout(self):
        """The file contains a @pipeline decorator with a timeout."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline",
                            "",
                            "@pipeline('Test pipeline', timeout=42)",
                            "def test_pipeline():",
                            "    pass",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "name": "Test pipeline",
                    "parameters": [],
                    "timeout": 42,
                    "function": None,
                    "tasks": [],
                },
            )

    def test_pipeline_with_bool(self):
        """The file contains a @pipeline decorator and a @parameter decorator with a bool."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "",
                            "@parameter('test_param', name='Test Param', type=bool, default=True, help='Param help')",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "name": "Test pipeline",
                    "function": None,
                    "tasks": [],
                    "parameters": [
                        {
                            "choices": None,
                            "multiple": False,
                            "code": "test_param",
                            "type": "bool",
                            "name": "Test Param",
                            "default": True,
                            "widget": None,
                            "connection": None,
                            "help": "Param help",
                            "required": True,
                        }
                    ],
                    "timeout": None,
                },
            )

    def test_pipeline_with_multiple_parameters(self):
        """The file contains a @pipeline decorator and multiple @parameter decorators."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "",
                            "@parameter('test_param', name='Test Param', type=int, default=42, help='Param help')",
                            "@parameter('test_param2', name='Test Param 2', type=str, choices=['a', 'b'], help='Param help 2')",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "name": "Test pipeline",
                    "function": None,
                    "tasks": [],
                    "parameters": [
                        {
                            "choices": None,
                            "multiple": False,
                            "code": "test_param",
                            "type": "int",
                            "name": "Test Param",
                            "default": 42,
                            "widget": None,
                            "connection": None,
                            "help": "Param help",
                            "required": True,
                        },
                        {
                            "choices": ["a", "b"],
                            "multiple": False,
                            "code": "test_param2",
                            "type": "str",
                            "name": "Test Param 2",
                            "default": None,
                            "widget": None,
                            "connection": None,
                            "help": "Param help 2",
                            "required": True,
                        },
                    ],
                    "timeout": None,
                },
            )

    def test_pipeline_with_unsupported_parameter(self):
        """The file contains a @pipeline decorator and a @parameter decorator with an unsupported type."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "",
                            "@parameter('test_param', name='Test Param', type=object, default=42, help='Param help')",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            with self.assertRaises(InvalidParameterError):
                get_pipeline(tmpdirname)

    def test_pipeline_with_connection_parameter_for_dhis2(self):
        """The file contains a @pipeline decorator and a @parameter decorator with a connection type."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "from openhexa.sdk.pipelines.widgets import DHIS2Widget",
                            "",
                            "@parameter('dhis_con', name='DHIS2 Connection', type=DHIS2Connection, required=True)",
                            "@parameter('data_element_ids', name='Data Elements id', type=str, widget=DHIS2Widget.ORG_UNITS, connection='dhis_con', required=True)",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.maxDiff = None
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "name": "Test pipeline",
                    "function": None,
                    "tasks": [],
                    "parameters": [
                        {
                            "code": "dhis_con",
                            "type": "dhis2",
                            "name": "DHIS2 Connection",
                            "default": None,
                            "multiple": False,
                            "choices": None,
                            "widget": None,
                            "connection": None,
                            "help": None,
                            "required": True,
                        },
                        {
                            "code": "data_element_ids",
                            "type": "str",
                            "name": "Data Elements id",
                            "widget": DHIS2Widget.ORG_UNITS.value,
                            "connection": "dhis_con",
                            "default": None,
                            "multiple": False,
                            "choices": None,
                            "help": None,
                            "required": True,
                        },
                    ],
                    "timeout": None,
                },
            )

    def test_pipeline_with_connection_parameter_for_iaso(self):
        """The file contains a @pipeline decorator and a @parameter decorator with a connection type."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "from openhexa.sdk.pipelines.widgets import IASOWidget",
                            "",
                            "@parameter('iaso_con', name='IASO Connection', type=IASOConnection, required=True)",
                            "@parameter('org_units', name='OrgUnits', type=int, widget=IASOWidget.IASO_ORG_UNITS, connection='iaso_con', required=True)",
                            "@parameter('projects', name='Projects', type=int, widget=IASOWidget.IASO_PROJECTS, connection='iaso_con', required=True)",
                            "@parameter('forms', name='Forms', type=int, widget=IASOWidget.IASO_FORMS, connection='iaso_con', required=True)",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline(org_units, forms, projects):",
                            "    pass",
                            "",
                        ]
                    )
                )
            pipeline = get_pipeline(tmpdirname)
            self.maxDiff = None
            self.assertEqual(
                pipeline.to_dict(),
                {
                    "name": "Test pipeline",
                    "function": None,
                    "tasks": [],
                    "parameters": [
                        {
                            "code": "iaso_con",
                            "type": "iaso",
                            "name": "IASO Connection",
                            "default": None,
                            "multiple": False,
                            "choices": None,
                            "widget": None,
                            "connection": None,
                            "help": None,
                            "required": True,
                        },
                        {
                            "code": "org_units",
                            "type": "int",
                            "name": "OrgUnits",
                            "widget": IASOWidget.IASO_ORG_UNITS.value,
                            "connection": "iaso_con",
                            "default": None,
                            "multiple": False,
                            "choices": None,
                            "help": None,
                            "required": True,
                        },
                        {
                            "code": "projects",
                            "type": "int",
                            "name": "Projects",
                            "widget": IASOWidget.IASO_PROJECTS.value,
                            "connection": "iaso_con",
                            "default": None,
                            "multiple": False,
                            "choices": None,
                            "help": None,
                            "required": True,
                        },
                        {
                            "code": "forms",
                            "type": "int",
                            "name": "Forms",
                            "widget": IASOWidget.IASO_FORMS.value,
                            "connection": "iaso_con",
                            "default": None,
                            "multiple": False,
                            "choices": None,
                            "help": None,
                            "required": True,
                        },
                    ],
                    "timeout": None,
                },
            )

    def test_pipeline_wit_wrong_connection_parameter(self):
        """The file contains a @pipeline decorator and a @parameter decorator with a non-existing connection type."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "from openhexa.sdk.pipelines.parameter import DHIS2Widget",
                            "",
                            "@parameter('dhis_con', name='DHIS2 Connection', type=DHIS2Connection, required=True)",
                            "@pipeline('Test pipeline')",
                            "@parameter('data_element_ids', name='Data Elements id', type=str, widget=DHIS2Widget.ORG_UNITS, connection='sds_con', required=True)",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            with self.assertRaises(InvalidParameterError):
                get_pipeline(tmpdirname)

    def test_pipeline_with_dhis2_widget_without_connection(self):
        """The file contains a @pipeline decorator and a @parameter decorator with a widget parameter field."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline, parameter",
                            "from openhexa.sdk.pipelines.parameter import DHIS2Widget",
                            "",
                            "@parameter('test_field_for_widget', name='Widget Param', type=str, widget=DHIS2Widget.ORG_UNITS, help='Param help')",
                            "@pipeline('Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            with self.assertRaises(InvalidParameterError):
                get_pipeline(tmpdirname)

    def test_pipeline_with_deprecated_code_argument_with_name(self):
        """The file contains a @pipeline decorator with the deprecated 'code' argument."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline",
                            "",
                            "@pipeline(code='test', name='Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            with patch("sys.stdout", new=io.StringIO()) as fake_stdout:
                pipeline = get_pipeline(tmpdirname)
                self.assertEqual(pipeline.to_dict()["name"], "Test pipeline")
                self.assertIn(
                    "The 'code' argument is deprecated and should not be used as a keyword.", fake_stdout.getvalue()
                )

    def test_pipeline_with_deprecated_code_as_arg(self):
        """The file contains a @pipeline decorator with the deprecated 'code' argument."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            with open(f"{tmpdirname}/pipeline.py", "w") as f:
                f.write(
                    "\n".join(
                        [
                            "from openhexa.sdk.pipelines import pipeline",
                            "",
                            "@pipeline('test', name='Test pipeline')",
                            "def test_pipeline():",
                            "    pass",
                            "",
                        ]
                    )
                )
            with patch("sys.stdout", new=io.StringIO()) as fake_stdout:
                pipeline = get_pipeline(tmpdirname)
                self.assertEqual(pipeline.to_dict()["name"], "Test pipeline")
                self.assertIn("Providing both 'code' and 'name' is deprecated.", fake_stdout.getvalue())



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel
from .enums import CreatePipelineFromTemplateVersionError


class CreatePipelineFromTemplateVersion(BaseModel):
    create_pipeline_from_template_version: (
        "CreatePipelineFromTemplateVersionCreatePipelineFromTemplateVersion"
    ) = Field(alias="createPipelineFromTemplateVersion")


class CreatePipelineFromTemplateVersionCreatePipelineFromTemplateVersion(BaseModel):
    success: bool
    errors: Optional[List[CreatePipelineFromTemplateVersionError]]
    pipeline: Optional[
        "CreatePipelineFromTemplateVersionCreatePipelineFromTemplateVersionPipeline"
    ]


class CreatePipelineFromTemplateVersionCreatePipelineFromTemplateVersionPipeline(
    BaseModel
):
    id: Any
    name: Optional[str]
    code: str


CreatePipelineFromTemplateVersion.model_rebuild()
CreatePipelineFromTemplateVersionCreatePipelineFromTemplateVersion.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel
from .enums import UpdateWorkspaceError


class UpdateWorkspace(BaseModel):
    update_workspace: "UpdateWorkspaceUpdateWorkspace" = Field(alias="updateWorkspace")


class UpdateWorkspaceUpdateWorkspace(BaseModel):
    success: bool
    workspace: Optional["UpdateWorkspaceUpdateWorkspaceWorkspace"]
    errors: List[UpdateWorkspaceError]


class UpdateWorkspaceUpdateWorkspaceWorkspace(BaseModel):
    slug: str
    name: str
    description: Optional[str]
    configuration: Optional[Any]
    countries: List["UpdateWorkspaceUpdateWorkspaceWorkspaceCountries"]


class UpdateWorkspaceUpdateWorkspaceWorkspaceCountries(BaseModel):
    code: str
    alpha_3: str = Field(alias="alpha3")
    name: str


UpdateWorkspace.model_rebuild()
UpdateWorkspaceUpdateWorkspace.model_rebuild()
UpdateWorkspaceUpdateWorkspaceWorkspace.model_rebuild()



"""User settings for the OpenHexa CLI."""

import logging
import os
from configparser import ConfigParser

import click

from openhexa.sdk.pipelines.log_level import LogLevel

CONFIGFILE_PATH = os.path.expanduser("~") + "/.openhexa.ini"


def _open_config():
    """Open the local settings file using configparser.

    A default settings file will be generated if the file does not exist.
    """
    config = ConfigParser()
    if os.path.exists(CONFIGFILE_PATH):
        config.read(CONFIGFILE_PATH)
    else:
        config.read_string(
            """
        [openhexa]
        url=https://api.openhexa.org

        [workspaces]
        """
        )
    return config


def _save_config(config: ConfigParser):
    """Save the provided configparser local settings to disk."""
    with open(CONFIGFILE_PATH, "w") as configfile:
        config.write(configfile)


class Settings:
    """Class that holds the user settings for the CLI."""

    _file_config: ConfigParser = None

    def __init__(self):
        self.refresh()

    @property
    def debug(self):
        """Return the debug flag from the settings file or environment variables."""
        return os.getenv("DEBUG") or os.getenv("HEXA_DEBUG")

    @property
    def api_url(self):
        """Return the API URL from the settings file or environment variables."""
        url_from_env = os.getenv("HEXA_API_URL") or os.getenv("HEXA_SERVER_URL")
        if url_from_env is None:
            url_from_env = self._file_config["openhexa"]["url"]
        return url_from_env.rstrip("/")

    @property
    def public_api_url(self):
        """Return the public API URL from the settings file or environment variables."""
        return self.api_url.replace("api", "app")

    @property
    def current_workspace(self):
        """Return the current workspace from the settings file or environment variables."""
        if os.getenv("HEXA_WORKSPACE"):
            return os.getenv("HEXA_WORKSPACE")
        try:
            return self._file_config["openhexa"]["current_workspace"]
        except KeyError:
            return None

    @property
    def workspaces(self):
        """Return the workspaces from the settings file."""
        return self._file_config["workspaces"]

    @property
    def log_level(self) -> LogLevel:
        """Return the log level from the environment variables."""
        return LogLevel.parse_log_level(os.getenv("HEXA_LOG_LEVEL"))

    def activate(self, workspace: str):
        """Set the current workspace in the settings file."""
        if workspace not in self.workspaces:
            raise ValueError(f"Workspace {workspace} does not exist.")
        self._file_config["openhexa"]["current_workspace"] = workspace
        self.save()

    def add_workspace(self, workspace: str, token: str, enabled: bool = True):
        """Add a workspace to the settings file."""
        self._file_config["workspaces"].update({workspace: token})
        if enabled:
            self._file_config["openhexa"]["current_workspace"] = workspace
        self.save()

    def remove_workspace(self, workspace: str):
        """Remove a workspace from the settings file."""
        del self._file_config["workspaces"][workspace]
        if self._file_config["openhexa"]["current_workspace"] == workspace:
            self._file_config["openhexa"]["current_workspace"] = None
        self.save()

    def set_api_url(self, url: str):
        """Set the API URL in the settings file."""
        self._file_config["openhexa"]["url"] = url
        self.save()

    @property
    def access_token(self):
        """Return the access token from the settings file or environment variables."""
        if os.getenv("HEXA_TOKEN"):
            return os.getenv("HEXA_TOKEN")
        return self._file_config["workspaces"].get(self.current_workspace)

    @property
    def last_version_check(self):
        """Return the last version check timestamp from the settings file."""
        val = self._file_config["openhexa"].get("last_version_check", None)
        if val is not None:
            return int(val)

    @last_version_check.setter
    def last_version_check(self, value: int):
        """Set the last version check timestamp in the settings file."""
        assert isinstance(value, int), "last_version_check must be an integer."
        self._file_config["openhexa"]["last_version_check"] = str(value)
        self.save()

    @property
    def last_breaking_change_check(self):
        """Return the last breaking change check timestamp from the settings file."""
        val = self._file_config["openhexa"].get("last_breaking_change_check", None)
        if val is not None:
            return int(val)

    @last_breaking_change_check.setter
    def last_breaking_change_check(self, value: int):
        """Set the last breaking change check timestamp in the settings file."""
        assert isinstance(value, int), "last_breaking_change_check must be an integer."
        self._file_config["openhexa"]["last_breaking_change_check"] = str(value)
        self.save()

    def save(self):
        """Save the settings to disk."""
        _save_config(self._file_config)
        self.refresh()

    def refresh(self):
        """Refresh the settings file."""
        self._file_config = _open_config()

    def __repr__(self):
        """Return a string representation of the object."""
        return f"<Config(url={self.api_url}, current_workspace={self.current_workspace}, debug={self.debug})>"


class ClickEchoHandler(logging.Handler):
    """Custom logging handler that uses click.echo to print logs."""

    def emit(self, record):
        """Emit the log record."""
        msg = self.format(record)
        click.echo(msg)


def setup_logging():
    """Set up the logging for the CLI."""
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    # Create a formatter
    formatter = logging.Formatter("%(message)s")

    # Create an instance of the custom handler
    click_handler = ClickEchoHandler()

    # Set the formatter for the handler
    click_handler.setFormatter(formatter)

    # Add the handler to the logger
    logger.addHandler(click_handler)


settings = Settings()



"""Workspaces package.

See https://github.com/BLSQ/openhexa/wiki/User-manual#about-workspaces for more information about OpenHEXA workspaces.
"""

from .current_workspace import CurrentWorkspace

# Once we deprecate the `python pipeline.py` command, we can enhance this to only load the workspace
# if we're in a pipeline/jupyter context
workspace = CurrentWorkspace()


__all__ = ["workspace"]



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel


class Organization(BaseModel):
    organization: Optional["OrganizationOrganization"]


class OrganizationOrganization(BaseModel):
    id: Any
    name: str
    short_name: Optional[str] = Field(alias="shortName")
    workspaces: "OrganizationOrganizationWorkspaces"
    permissions: "OrganizationOrganizationPermissions"


class OrganizationOrganizationWorkspaces(BaseModel):
    items: List["OrganizationOrganizationWorkspacesItems"]


class OrganizationOrganizationWorkspacesItems(BaseModel):
    slug: str
    name: str
    countries: List["OrganizationOrganizationWorkspacesItemsCountries"]


class OrganizationOrganizationWorkspacesItemsCountries(BaseModel):
    code: str


class OrganizationOrganizationPermissions(BaseModel):
    create_workspace: bool = Field(alias="createWorkspace")
    archive_workspace: bool = Field(alias="archiveWorkspace")


Organization.model_rebuild()
OrganizationOrganization.model_rebuild()
OrganizationOrganizationWorkspaces.model_rebuild()
OrganizationOrganizationWorkspacesItems.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/cli/graphql/queries.graphql

from typing import Optional

from .base_model import BaseModel


class GetWorkspace(BaseModel):
    workspace: Optional["GetWorkspaceWorkspace"]


class GetWorkspaceWorkspace(BaseModel):
    name: str
    slug: str


GetWorkspace.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import Any, List

from .base_model import BaseModel


class Organizations(BaseModel):
    organizations: List["OrganizationsOrganizations"]


class OrganizationsOrganizations(BaseModel):
    id: Any
    name: str
    workspaces: "OrganizationsOrganizationsWorkspaces"


class OrganizationsOrganizationsWorkspaces(BaseModel):
    items: List["OrganizationsOrganizationsWorkspacesItems"]


class OrganizationsOrganizationsWorkspacesItems(BaseModel):
    slug: str
    name: str


Organizations.model_rebuild()
OrganizationsOrganizations.model_rebuild()
OrganizationsOrganizationsWorkspaces.model_rebuild()



"""Dataset test module."""

import os
from unittest import TestCase
from unittest.mock import patch

from httmock import HTTMock, all_requests, response

from openhexa.sdk.datasets import Dataset
from openhexa.sdk.workspaces import workspace


class DatasetTest(TestCase):
    """Dataset test class."""

    @patch.dict(
        os.environ,
        {
            "HEXA_WORKSPACE": "workspace-slug",
            "HEXA_TOKEN": "token",
            "HEXA_SERVER_URL": "http://server",
        },
    )
    def test_create_dataset_ok(self):
        """Ensure that datasets can be created."""

        @all_requests
        def graphql_responses(url, request):
            if b"createDataset" in request.body:
                return response(
                    200,
                    {
                        "data": {
                            "createDataset": {
                                "success": True,
                                "dataset": {
                                    "id": "dataset_id",
                                    "slug": "my-dataset",
                                    "name": "My Dataset",
                                    "description": "My Dataset description",
                                    "createdAt": "2021-01-01T00:00:00.000Z",
                                },
                            }
                        }
                    },
                    request=request,
                )
            elif b"getDataset" in request.body:
                return response(
                    200,
                    {
                        "data": {
                            "datasetLinkBySlug": {
                                "id": "link_id",
                                "workspace": "WORKSPACE",
                                "dataset": {
                                    "id": "dataset_id",
                                    "slug": "my-dataset",
                                    "name": "My Dataset",
                                    "description": "My Dataset description",
                                    "createdAt": "2021-01-01T00:00:00.000Z",
                                },
                            }
                        }
                    },
                    request=request,
                )

        with HTTMock(graphql_responses):
            dataset = workspace.create_dataset("My Dataset", "My Dataset description")
            self.assertEqual(dataset.id, "dataset_id")
            self.assertEqual(dataset.slug, "my-dataset")

    @patch.dict(
        os.environ,
        {
            "HEXA_WORKSPACE": "workspace-slug",
            "HEXA_TOKEN": "token",
            "HEXA_SERVER_URL": "server",
        },
    )
    @patch("openhexa.sdk.datasets.dataset.graphql")
    def test_create_dataset_version(self, mock_graphql):
        """Ensure that dataset versions can be created."""
        d = Dataset(
            id="id",
            slug="my-dataset",
            name="My Dataset",
            description="My Dataset description",
        )

        mock_graphql.return_value = {
            "createDatasetVersion": {
                "success": True,
                "version": {
                    "id": "<newVersionId>",
                    "name": "Second version",
                    "description": "Description",
                    "createdAt": "2021-01-01T00:00:00.000Z",
                },
            }
        }

        v = d.create_version("Second version")
        self.assertEqual(v.id, "<newVersionId>")
        v = d.create_version("Second version")
        self.assertEqual(v.id, "<newVersionId>")



"""Pipeline run module."""

import datetime
import errno
import os

from openhexa.sdk.pipelines.log_level import LogLevel
from openhexa.sdk.utils import Environment, get_environment, graphql
from openhexa.sdk.workspaces import workspace


class CurrentRun:
    """Represents the current run of a pipeline.

    CurrentRun instances allow pipeline developers to interact with the OpenHEXA backend, by sending messages and
    adding outputs that will be available through the web interface.
    """

    @property
    def _connected(self):
        return "HEXA_SERVER_URL" in os.environ

    def id(self):
        """Exposes pipeline run id."""
        if "HEXA_RUN_ID" in os.environ:
            return os.environ["HEXA_RUN_ID"]
        else:
            raise RuntimeError("Current run ID is not set. Make sure you are running this in a pipeline context.")

    def add_file_output(self, path: str):
        """Record a run output for a file creation operation.

        This output will be visible in the web interface, on the pipeline run page.
        """
        stripped_path = path.replace(workspace.files_path, "")
        name = stripped_path.strip("/")
        if self._connected:
            res = graphql(
                """
                mutation addPipelineOutput ($input: AddPipelineOutputInput!) {
                    addPipelineOutput(input: $input) { success errors }
                }""",
                {
                    "input": {
                        "uri": f"gs://{os.environ['WORKSPACE_BUCKET_NAME']}{stripped_path}",
                        "type": "file",
                        "name": name,
                    }
                },
            )
            if not res["addPipelineOutput"]["success"]:
                if "FILE_NOT_FOUND" in res["addPipelineOutput"]["errors"]:
                    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

                raise Exception(res["addPipelineOutput"]["errors"])
        else:
            print(f"Sending output with path {stripped_path}")

    def add_database_output(self, table_name: str):
        """Record a run output for a database operation.

        This output will be visible in the web interface, on the pipeline run page.
        """
        if self._connected:
            res = graphql(
                """
                mutation addPipelineOutput ($input: AddPipelineOutputInput!) {
                    addPipelineOutput(input: $input) { success errors }
                }""",
                {
                    "input": {
                        "uri": f"postgresql://{workspace.database_host}/{workspace.database_name}/{table_name}",
                        "type": "db",
                        "name": table_name,
                    }
                },
            )
            if not res["addPipelineOutput"]["success"]:
                if "TABLE_NOT_FOUND" in res["addPipelineOutput"]["errors"]:
                    raise Exception(f"{table_name} doesn't exist in workspace {workspace.slug}")
        else:
            print(f"Sending output with table_name {table_name}")

    def log_debug(self, message: str):
        """Log a message with the DEBUG level."""
        self._log_message(LogLevel.DEBUG, message)

    def log_info(self, message: str):
        """Log a message with the INFO level."""
        self._log_message(LogLevel.INFO, message)

    def log_warning(self, message: str):
        """Log a message with the WARNING level."""
        self._log_message(LogLevel.WARNING, message)

    def log_error(self, message: str):
        """Log a message with the ERROR level."""
        self._log_message(LogLevel.ERROR, message)

    def log_critical(self, message: str):
        """Log a message with the CRITICAL level."""
        self._log_message(LogLevel.CRITICAL, message)

    def _log_message(
        self,
        log_level: LogLevel,
        message: str,
    ):
        from openhexa.cli.settings import settings

        if log_level < settings.log_level:  # Ignore messages with lower log level than the settings
            return
        if self._connected:
            graphql(
                """
                mutation logPipelineMessage ($input: LogPipelineMessageInput!) {
                    logPipelineMessage(input: $input) { success errors }
                }""",
                {"input": {"priority": log_level.name, "message": str(message)}},
            )
        else:
            now = datetime.datetime.now(tz=datetime.UTC).replace(microsecond=0).isoformat()
            print(now, log_level.name, message)


if get_environment() == Environment.CLOUD_JUPYTER:
    current_run = None
else:
    current_run = CurrentRun()



"""Dataset-related classes and functions.

See https://github.com/BLSQ/openhexa/wiki/User-manual#datasets and
https://github.com/BLSQ/openhexa/wiki/Using-the-OpenHEXA-SDK#working-with-datasets for more information about datasets.
"""

import mimetypes
import typing
from os import PathLike
from pathlib import Path

import requests

from openhexa.sdk.utils import Iterator, Page, graphql, read_content


class DatasetFile:
    """Represent a single file within a dataset. Files are attached to dataset through versions."""

    _download_url = None
    version = None

    def __init__(
        self,
        version: any,
        id: str,
        uri: str,
        filename: str,
        content_type: str,
        created_at: str,
    ):
        self.version = version
        self.id = id
        self.uri = uri
        self.filename = filename
        self.content_type = content_type
        self.created_at = created_at

    def read(self):
        """Download the file content and return it."""
        response = requests.get(self.download_url, stream=True)
        response.raise_for_status()
        return response.content

    @property
    def download_url(self):
        """Build and return a pre-signed URL for the file."""
        if self._download_url is None:
            response = graphql(
                """
                mutation getDownloadUrl($input: PrepareVersionFileDownloadInput!) {
                    prepareVersionFileDownload(input: $input) {
                        downloadUrl
                        success
                        errors
                    }
                }
            """,
                {"input": {"fileId": self.id}},
            )
            if response["prepareVersionFileDownload"]["success"] is False:
                raise Exception(response["prepareVersionFileDownload"]["errors"])
            self._download_url = response["prepareVersionFileDownload"]["downloadUrl"]
        return self._download_url

    def __repr__(self) -> str:
        """Safe representation of the dataset file."""
        return f"<DatasetFile id={self.id} filename={self.filename}>"


class VersionsIterator(Iterator):
    """Custom iterator class to iterate versions using our GraphQL API."""

    def __init__(self, dataset: any, per_page: int = 10):
        super().__init__(per_page=per_page)

        self.item_to_value = lambda x: DatasetVersion(
            dataset=dataset, id=x["id"], name=x["name"], created_at=x["createdAt"]
        )
        self.dataset = dataset
        self.has_next_page = True

    def _next_page(self):
        if not self.has_next_page:
            return None

        res = graphql(
            """
            query getDatasetVersions($datasetId: ID!, $page: Int!, $perPage: Int) {
                dataset (id: $datasetId) {
                    versions (page: $page, perPage: $perPage) {
                        items {
                            id
                            name
                            createdAt
                        }
                        totalPages
                    }
                }
            }
            """,
            {
                "datasetId": self.dataset.id,
                "page": self.page_number + 1,
                "perPage": self.per_page,
            },
        )
        if res["dataset"] is None:
            raise ValueError(f"Dataset {self.dataset.id} does not exist")

        if res["dataset"]["versions"]["totalPages"] == self.page_number + 1:
            self.has_next_page = False

        return Page(
            parent=self,
            items=res["dataset"]["versions"]["items"],
            item_to_value=self.item_to_value,
        )


class VersionFilesIterator(Iterator):
    """Custom iterator class to iterate version files using our GraphQL API."""

    def __init__(self, version: any, per_page: int = 20):
        super().__init__(per_page=per_page)
        self.item_to_value = lambda x: DatasetFile(
            version=version,
            id=x["id"],
            uri=x["uri"],
            filename=x["filename"],
            content_type=x["contentType"],
            created_at=x["createdAt"],
        )

        self.version = version
        self.has_next_page = True

    def _next_page(self):
        if not self.has_next_page:
            return None

        res = graphql(
            """
            query getDatasetFiles($versionId: ID!, $page: Int!, $perPage: Int) {
                datasetVersion (id: $versionId) {
                    files (page: $page, perPage: $perPage) {
                        items {
                            id
                            uri
                            filename
                            contentType
                            createdAt
                        }
                        totalPages
                    }
                }
            }
            """,
            {
                "versionId": self.version.id,
                "page": self.page_number + 1,
                "perPage": self.per_page,
            },
        )
        if res["datasetVersion"] is None:
            raise ValueError(f"Dataset version {self.version.id} does not exist")

        if res["datasetVersion"]["files"]["totalPages"] == self.page_number + 1:
            self.has_next_page = False

        return Page(
            parent=self,
            items=res["datasetVersion"]["files"]["items"],
            item_to_value=self.item_to_value,
        )


class DatasetVersion:
    """Dataset files are not directly attached to a dataset, but rather to a version."""

    dataset = None

    def __init__(self, dataset: any, id: str, name: str, created_at: str):
        self.id = id
        self.name = name
        self.dataset = dataset
        self.created_at = created_at

    @property
    def files(self):
        """Build and return an iterator of files for this version."""
        if self.id is None:
            raise ValueError("This dataset version does not have an id.")
        return VersionFilesIterator(version=self, per_page=50)

    def get_file(self, filename: str) -> DatasetFile:
        """Get a file by name."""
        data = graphql(
            """
            query getDatasetFile($versionId: ID!, $filename: String!) {
                datasetVersion(id: $versionId) {
                    fileByName(name: $filename) {
                        id
                        uri
                        filename
                        contentType
                        createdAt
                    }
                }
            }
            """,
            {"versionId": self.id, "filename": filename},
        )

        file = data["datasetVersion"]["fileByName"]
        if file is None:
            raise FileExistsError(f"The file {filename} does not exist for version {self}")

        return DatasetFile(
            version=self,
            id=file["id"],
            uri=file["uri"],
            filename=file["filename"],
            content_type=file["contentType"],
            created_at=file["createdAt"],
        )

    def add_file(
        self,
        source: str | PathLike[str] | typing.IO | bytes,
        filename: str | None = None,
    ) -> DatasetFile:
        """Create a new dataset file and add it to the dataset version."""
        mime_type = None
        if isinstance(source, (str | PathLike)):
            path = Path(source)
            filename = path.name if filename is None else filename
            mime_type, _ = mimetypes.guess_type(path)
        else:
            if filename is None:
                raise ValueError("A file name is required when you pass a buffer")

        if mime_type is None:
            mime_type = "application/octet-stream"

        upload_url_result = graphql(
            """
            mutation generateDatasetUploadUrl ($input: GenerateDatasetUploadUrlInput!) {
                generateDatasetUploadUrl(input: $input) {
                    uploadUrl
                    success
                    errors
                }
            }
            """,
            {"input": {"versionId": self.id, "contentType": mime_type, "uri": filename}},
        )
        if upload_url_result["generateDatasetUploadUrl"]["success"] is False:
            errors = upload_url_result["generateDatasetUploadUrl"]["errors"]
            self.raise_upload_exception(errors)

        upload_url = upload_url_result["generateDatasetUploadUrl"]["uploadUrl"]
        with read_content(source) as content:
            response = requests.put(upload_url, data=content, headers={"Content-Type": mime_type})
        response.raise_for_status()

        data = graphql(
            """
                mutation CreateDatasetVersionFile ($input: CreateDatasetVersionFileInput!) {
                    createDatasetVersionFile(input: $input) {
                        file {
                            id
                            filename
                            uri
                            contentType
                            createdAt
                        }
                        success
                        errors
                    }
                }
        """,
            {
                "input": {
                    "versionId": self.id,
                    "contentType": mime_type,
                    "uri": filename,
                }
            },
        )

        if data["createDatasetVersionFile"]["success"] is False:
            errors = data["createDatasetVersionFile"]["errors"]
            self.raise_dataset_file_creation_exception(errors)
        result = data["createDatasetVersionFile"]

        return DatasetFile(
            version=self,
            id=result["file"]["id"],
            filename=result["file"]["filename"],
            content_type=result["file"]["contentType"],
            uri=result["file"]["uri"],
            created_at=result["file"]["createdAt"],
        )

    def exists(self, objectKey: str):
        """
        Check if an object with the specified key exists.

        This method checks if an object identified by `objectKey` exists in the dataset.

        Args:
            objectKey (str): The key of the object to check for existence.

        Returns
        -------
            bool: True if the object exists, False otherwise.
        """
        data = graphql(
            """
            query getDatasetFile($versionId: ID!, $filename: String!) {
                datasetVersion(id: $versionId) {
                    fileByName(name: $filename) {
                        id
                        uri
                        filename
                        contentType
                        createdAt
                    }
                }
            }
            """,
            {"versionId": self.id, "filename": objectKey},
        )

        return data["datasetVersion"]["fileByName"] is not None

    def raise_upload_exception(self, errors):
        """Raise an exception if an error occurs on upload."""
        if "LOCKED_VERSION" in errors:
            raise ValueError("This dataset version is locked. You can only add files to the latest version")
        elif "ALREADY_EXISTS" in errors:
            raise ValueError("A file with this name already exists in this dataset version")
        else:
            raise Exception(errors)

    def raise_dataset_file_creation_exception(self, errors):
        """Raise an exception if an error occurs on upload."""
        if "LOCKED_VERSION" in errors:
            raise ValueError("This dataset version is locked. You can only add files to the latest version")
        elif "PERMISSION_DENIED" in errors:
            raise ValueError("You do not have permission to add files to this dataset version")
        elif "ALREADY_EXISTS" in errors:
            raise ValueError("A file with this name already exists in this dataset version")
        else:
            raise Exception(errors)


class Dataset:
    """Datasets are versioned, documented files.

    See https://github.com/BLSQ/openhexa/wiki/Using-the-OpenHEXA-SDK#working-with-datasets for more information.
    """

    _latest_version = None

    def __init__(
        self,
        id: str,
        slug: str,
        name: str,
        description: str,
    ):
        self.id = id
        self.slug = slug
        self.name = name
        self.description = description

    def create_version(self, name: typing.Any) -> DatasetVersion:
        """Build a dataset version, save it and return it."""
        response = graphql(
            """
            mutation createDatasetVersion($input: CreateDatasetVersionInput!) {
                createDatasetVersion(input: $input) {
                    version {
                        id
                        name
                        description
                        createdAt
                    }
                    errors
                    success
                }
            }
        """,
            {
                "input": {
                    "datasetId": self.id,
                    "name": str(name),
                }
            },
        )
        data = response["createDatasetVersion"]
        if data["success"] is False:
            if "DUPLICATE_NAME" in data["errors"]:
                raise ValueError("A dataset version with this name already exists")
            raise Exception(data["errors"])

        version = data["version"]
        self._latest_version = DatasetVersion(
            dataset=self,
            id=version["id"],
            name=version["name"],
            created_at=version["createdAt"],
        )

        return self.latest_version

    @property
    def latest_version(self) -> DatasetVersion | None:
        """Return the latest version, if any.

        This property method will query the backend to try to fetch the latest version.
        """
        if self._latest_version is None:
            data = graphql(
                """
                query getLatestVersion($datasetId: ID!) { 
                    dataset(id: $datasetId) {
                        latestVersion {
                            id
                            name
                            createdAt
                        }
                    }
                }
            """,
                {"datasetId": self.id},
            )

            if data["dataset"]["latestVersion"] is None:
                self._latest_version = None
            else:
                self._latest_version = DatasetVersion(
                    dataset=self,
                    id=data["dataset"]["latestVersion"]["id"],
                    name=data["dataset"]["latestVersion"]["name"],
                    created_at=data["dataset"]["latestVersion"]["createdAt"],
                )

        return self._latest_version

    @property
    def versions(self) -> VersionsIterator:
        """Build and return an iterator for versions."""
        return VersionsIterator(dataset=self, per_page=10)

    def __repr__(self) -> str:
        """Safe representation of the dataset."""
        return f"<Dataset slug={self.slug} id={self.id}>"


class FileNotFound(Exception):
    """Raised whenever an attempt is made to get a file that does not exist."""



"""Exceptions for the pipelines module."""


class PipelineNotFound(Exception):
    """Errors related to pipeline not found in a file."""

    pass


class InvalidParameterError(Exception):
    """Raised whenever parameter options (usually passed to the @parameter decorator) are invalid."""

    pass


class ParameterValueError(Exception):
    """Raised whenever values for a parameter provided for a pipeline run are invalid."""

    pass



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List

from pydantic import Field

from .base_model import BaseModel
from .enums import AddToFavoritesError


class AddWebappToFavorites(BaseModel):
    add_to_favorites: "AddWebappToFavoritesAddToFavorites" = Field(
        alias="addToFavorites"
    )


class AddWebappToFavoritesAddToFavorites(BaseModel):
    success: bool
    errors: List[AddToFavoritesError]


AddWebappToFavorites.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List

from pydantic import Field

from .base_model import BaseModel
from .enums import DeleteWebappError


class DeleteWebapp(BaseModel):
    delete_webapp: "DeleteWebappDeleteWebapp" = Field(alias="deleteWebapp")


class DeleteWebappDeleteWebapp(BaseModel):
    success: bool
    errors: List[DeleteWebappError]


DeleteWebapp.model_rebuild()



"""Workspace-related classes and functions.

See https://github.com/BLSQ/openhexa/wiki/User-manual#about-workspaces for more information.
"""

import os
from dataclasses import fields, make_dataclass
from warnings import warn

from openhexa.graphql.graphql_client import WorkspaceWorkspaceCountries
from openhexa.graphql.graphql_client.input_types import UpdateWorkspaceInput
from openhexa.utils import stringcase

from ..datasets import Dataset
from ..utils import OpenHexaClient, graphql
from .connection import (
    ConnectionClasses,
    CustomConnection,
    DHIS2Connection,
    GCSConnection,
    IASOConnection,
    PostgreSQLConnection,
    S3Connection,
)


class WorkspaceConfigError(Exception):
    """Raised whenever the system cannot find an environment variable required to configure the current workspace."""

    pass


class ConnectionDoesNotExist(Exception):
    """Raised whenever an attempt is made to get a connection through an invalid identifier."""

    pass


class CurrentWorkspace:
    """Represents the currently configured OpenHEXA workspace, with its filesystem, database and connections."""

    @property
    def _connected(self):
        return "HEXA_SERVER_URL" in os.environ

    @property
    def _token(self) -> str:
        try:
            return os.environ["HEXA_TOKEN"]
        except KeyError:
            raise WorkspaceConfigError("The workspace token is not available in this environment.")

    @property
    def slug(self) -> str:
        """The unique slug of the workspace.

        Slugs are used to identify the workspace.
        """
        try:
            return os.environ["HEXA_WORKSPACE"]
        except KeyError:
            raise WorkspaceConfigError("The workspace slug is not available in this environment.")

    @property
    def countries(self) -> list[WorkspaceWorkspaceCountries]:
        """The countries of the workspace."""
        try:
            return OpenHexaClient().workspace(slug=self.slug).countries
        except KeyError:
            raise WorkspaceConfigError("The workspace countries are not available in this environment.")

    @property
    def configuration(self) -> dict[str, str | dict] | None:
        """The workspace configuration as a dictionary.

        Returns a dictionary containing workspace configuration as key-value pairs,
        where keys are strings and values can be either strings or JSON objects.
        Returns None if no configuration is available or if not connected to the API.
        """
        if not self._connected:
            return None
        return OpenHexaClient().workspace(slug=self.slug).configuration

    @configuration.setter
    def configuration(self, value: dict[str, str | dict]) -> None:
        """Set the workspace configuration.

        Parameters
        ----------
        value : dict[str, str | dict]
            The configuration dictionary to set for the workspace.
            Keys must be strings and values can be strings or JSON objects.

        Raises
        ------
        WorkspaceConfigError
            If not connected to the API or if the update fails.

        Examples
        --------
        >>> workspace.configuration = {
        ...     "api_url": "https://api.example.com",
        ...     "debug_mode": "true",
        ...     "SNT_CONFIG": "VERY_GOOD_CONFIG"
        ... }
        """
        if not self._connected:
            raise WorkspaceConfigError("Cannot update configuration: not connected to the API.")

        try:
            client = OpenHexaClient()

            input_data = UpdateWorkspaceInput(slug=self.slug, configuration=value)
            result = client.update_workspace(input=input_data)

            if not result.success:
                raise WorkspaceConfigError("Failed to update workspace configuration.")

        except Exception as e:
            if isinstance(e, WorkspaceConfigError):
                raise
            raise WorkspaceConfigError(f"Failed to update workspace configuration: {str(e)}")

    @property
    def database_host(self) -> str:
        """The workspace database host."""
        try:
            return os.environ["WORKSPACE_DATABASE_HOST"]
        except KeyError:
            raise WorkspaceConfigError(
                "No database has been configured. Did you forget to provide a database entry"
                " in your workspace.yaml file?"
            )

    @property
    def database_username(self) -> str:
        """The workspace database username."""
        try:
            return os.environ["WORKSPACE_DATABASE_USERNAME"]
        except KeyError:
            raise WorkspaceConfigError(
                "No database has been configured. Did you forget to provide a database entry"
                " in your workspace.yaml file?"
            )

    @property
    def database_password(self):
        """The workspace database password."""
        try:
            return os.environ.get("WORKSPACE_DATABASE_PASSWORD")
        except KeyError:
            raise WorkspaceConfigError(
                "No database has been configured. Did you forget to provide a database entry"
                " in your workspace.yaml file?"
            )

    @property
    def database_name(self):
        """The workspace database name."""
        try:
            return os.environ["WORKSPACE_DATABASE_DB_NAME"]
        except KeyError:
            raise WorkspaceConfigError(
                "No database has been configured. Did you forget to provide a database entry"
                " in your workspace.yaml file?"
            )

    @property
    def database_port(self):
        """The workspace database port."""
        try:
            return int(os.environ["WORKSPACE_DATABASE_PORT"])
        except KeyError:
            raise WorkspaceConfigError(
                "No database has been configured. Did you forget to provide a database entry"
                " in your workspace.yaml file?"
            )

    @property
    def database_url(self):
        """The workspace database URL.

        The URL follows the official PostgreSQL specification.
        (See https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-CONNSTRING for more information)
        """
        return (
            f"postgresql://{self.database_username}:{self.database_password}"
            f"@{self.database_host}:{self.database_port}/{self.database_name}"
        )

    @property
    def files_path(self) -> str:
        """The base path to the filesystem, without trailing slash.

        Examples
        --------
        >>> f"{workspace.files_path}/some/path"
        /home/hexa/workspace/some/path
        """
        # FIXME: This is a hack to make the SDK work in the context of the `python pipeline.py` command.
        # We can remove this once we deprecate this way of running pipelines and only use /home/hexa/workspace
        return os.environ.get("WORKSPACE_FILES_PATH", "/home/hexa/workspace")

    @property
    def tmp_path(self) -> str:
        """The base path to the tmp directory, without trailing slash.

        Examples
        --------
        >>> f"{workspace.tmp_path}/some/path"
        /home/hexa/tmp/some/path
        """
        # FIXME: This is a hack to make the SDK work in the context of the `python pipeline.py` command.
        # We can remove this once we deprecate this way of running pipelines
        return os.environ["WORKSPACE_TMP_PATH"] if "WORKSPACE_TMP_PATH" in os.environ else "/home/hexa/tmp"

    def _get_local_connection_fields(self, env_variable_prefix: str):
        connection_fields = {}
        connection_type = os.getenv(env_variable_prefix).upper()

        # Get fields for the connection type
        _fields = fields(ConnectionClasses[connection_type])

        if _fields:
            for field in _fields:
                env_var = f"{env_variable_prefix}_{field.name.upper()}"
                connection_fields[field.name] = os.getenv(env_var)
        else:
            #  custom connections
            prefix = f"{env_variable_prefix}_"
            connection_fields = {
                key[len(prefix) :].lower(): val for key, val in os.environ.items() if key.startswith(prefix)
            }

        # need to map the correct name for s3 and postgres connection to ensure compatibility
        # with the one coming from the API
        if connection_type == "S3":
            connection_fields.pop("secret_access_key")
            connection_fields["access_key_secret"] = os.getenv(f"{env_variable_prefix}_ACCESS_KEY_SECRET")
        if connection_type == "POSTGRESQL":
            connection_fields.pop("database_name")
            connection_fields["db_name"] = os.getenv(f"{env_variable_prefix}_DB_NAME")

        return connection_fields

    def get_connection_from_api(self, identifier: str) -> tuple[dict[str, str], str] | None:
        """Get a connection by its identifier from the OpenHEXA API."""
        connection_fields: dict[str, str] = {}
        connection = OpenHexaClient().get_connection(workspace_slug=self.slug, connection_slug=identifier.lower())
        if not connection:
            return None
        for f in connection.fields:
            connection_fields[f.code] = f.value
        connection_type = connection.type.upper()
        return connection_fields, connection_type

    def get_connection_from_env(self, identifier: str) -> tuple[dict[str, str], str] | None:
        """Get a connection by its identifier from the environment variables."""
        env_variable_prefix = stringcase.constcase(identifier.lower())
        try:
            connection_type = os.environ[f"{env_variable_prefix}"].upper()
            connection_fields = self._get_local_connection_fields(env_variable_prefix)
            return connection_fields, connection_type
        except KeyError:
            return None

    def get_connection(
        self, identifier: str
    ) -> (
        DHIS2Connection | PostgreSQLConnection | IASOConnection | S3Connection | GCSConnection | CustomConnection | None
    ):
        """Get a connection by its identifier.

        Parameters
        ----------
        identifier : str
            The identifier of the connection in the OpenHEXA backend

        Returns
        -------
        Connection
            The connection

        Raises
        ------
        ValueError
            If the connection does not exist
        """
        connection = self.get_connection_from_env(identifier)
        if not connection and self._connected:
            connection = self.get_connection_from_api(identifier)

        if not connection:
            raise ValueError(f"Connection {identifier} does not exist.")

        connection_fields, connection_type = connection

        # In connected mode (API call) the secret_access_key field and db_name name are
        # different from the offline ones
        if connection_type == "S3":
            secret_access_key = connection_fields.pop("access_key_secret")
            return S3Connection(secret_access_key=secret_access_key, **connection_fields)

        if connection_type == "POSTGRESQL":
            db_name = connection_fields.pop("db_name")
            port = int(connection_fields.pop("port"))
            return PostgreSQLConnection(
                database_name=db_name,
                port=port,
                **connection_fields,
            )

        if connection_type == "CUSTOM":
            dataclass = make_dataclass(
                stringcase.pascalcase(identifier),
                connection_fields.keys(),
                bases=(CustomConnection,),
                repr=False,
            )
            return dataclass(**connection_fields)

        return ConnectionClasses[connection_type](**connection_fields)

    def dhis2_connection(self, identifier: str = None, slug: str = None) -> DHIS2Connection:
        """Get a DHIS2 connection by its identifier.

        Parameters
        ----------
        identifier : str
            The identifier of the connection in the OpenHEXA backend
        slug : str
            Deprecated, same as identifier
        """
        identifier = identifier or slug
        if slug is not None:
            warn(
                "'slug' is deprecated. Use 'identifier' instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        try:
            connection = self.get_connection(identifier)
            assert isinstance(connection, DHIS2Connection), "Connection is not a DHIS2Connection"
            return connection

        except ValueError:
            raise ConnectionDoesNotExist(f'No DHIS2 connection for "{identifier}"')

    def postgresql_connection(self, identifier: str = None, slug: str = None) -> PostgreSQLConnection:
        """Get a PostgreSQL connection by its identifier.

        Parameters
        ----------
        identifier : str
            The identifier of the connection in the OpenHEXA backend
        slug : str
            Deprecated, same as identifier
        """
        identifier = identifier or slug
        if slug is not None:
            warn(
                "'slug' is deprecated. Use 'identifier' instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        try:
            connection = self.get_connection(identifier)
            assert isinstance(connection, PostgreSQLConnection), "Connection is not a PostgreSQLConnection"

            return connection
        except ValueError:
            raise ConnectionDoesNotExist(f'No PostgreSQL connection for "{identifier}"')

    def s3_connection(self, identifier: str = None, slug: str = None) -> S3Connection:
        """Get an AWS S3 connection by its identifier.

        Parameters
        ----------
        identifier : str
            The identifier of the connection in the OpenHEXA backend
        slug : str
            Deprecated, same as identifier
        """
        identifier = identifier or slug
        if slug is not None:
            warn(
                "'slug' is deprecated. Use 'identifier' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        try:
            connection = self.get_connection(identifier)
            assert isinstance(connection, S3Connection), "Connection is not a S3Connection"

            return connection
        except ValueError:
            raise ConnectionDoesNotExist(f'No S3 connection for "{identifier}"')

    def gcs_connection(self, identifier: str = None, slug: str = None) -> GCSConnection:
        """Get a Google Cloud Storage connection by its identifier.

        Parameters
        ----------
        identifier : str
            The identifier of the connection in the OpenHEXA backend
        slug : str
            Deprecated, same as identifier
        """
        identifier = identifier or slug
        if slug is not None:
            warn(
                "'slug' is deprecated. Use 'identifier' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        try:
            connection = self.get_connection(identifier)
            assert isinstance(connection, GCSConnection), "Connection is not a GCSConnection"

            return connection
        except ValueError:
            raise ConnectionDoesNotExist(f'No GCS connection for "{identifier}"')

    def iaso_connection(self, identifier: str = None, slug: str = None) -> IASOConnection:
        """Get a IASO connection by it identifier.

        Parameters
        ----------
        identifier : str
            The identifier of the connection in the OpenHEXA backend
        slug : str
            Deprecated, same as identifier
        """
        identifier = identifier or slug
        if slug is not None:
            warn(
                "'slug' is deprecated. Use 'identifier' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        try:
            connection = self.get_connection(identifier)
            assert isinstance(connection, IASOConnection), "Connection is not a IASOConnection"

            return connection
        except ValueError:
            raise ConnectionDoesNotExist(f'No IASO connection for "{identifier}"')

    def custom_connection(self, identifier: str = None, slug: str = None) -> CustomConnection:
        """Get a custom connection by its identifier.

        Parameters
        ----------
        identifier : str
            The identifier of the connection in the OpenHEXA backend
        slug : str
            Deprecated, same as identifier
        """
        identifier = identifier or slug
        if slug is not None:
            warn(
                "'slug' is deprecated. Use 'identifier' instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        try:
            connection = self.get_connection(identifier)
            assert isinstance(connection, CustomConnection), "Connection is not a CustomConnection"
            return connection
        except ValueError:
            raise ConnectionDoesNotExist(f'No Custom connection for "{identifier}"')

    def create_dataset(self, name: str, description: str):
        """Create a new dataset.

        Parameters
        ----------
        name: str
            The name of the dataset
        description: str
            The description of the dataset

        Returns
        -------
        Dataset
            The created dataset

        Raises
        ------
        ValueError
            If the dataset could not be created
        """
        rsp = graphql(
            """
        mutation createDataset($input: CreateDatasetInput!) {
            createDataset(input: $input) {
                success
                errors
                dataset {
                    slug
                }
            }
        }
        """,
            {
                "input": {
                    "workspaceSlug": self.slug,
                    "name": name,
                    "description": description,
                }
            },
        )

        if rsp["createDataset"]["success"] is False:
            raise ValueError(rsp["createDataset"]["errors"][0])

        identifier = rsp["createDataset"]["dataset"]["slug"]
        return self.get_dataset(identifier)

    def get_dataset(self, identifier: str) -> Dataset:
        """Get a dataset by its identifier.

        Parameters
        ----------
        identifier : str
            The identifier of the dataset in the OpenHEXA backend

        Returns
        -------
        Dataset
            The dataset

        Raises
        ------
        ValueError
            If the dataset does not exist
        """
        response = graphql(
            """
            query getDataset($datasetSlug: String!, $workspaceSlug: String!) {
                datasetLinkBySlug(datasetSlug: $datasetSlug, workspaceSlug: $workspaceSlug) {
                    id
                    workspace {
                        slug
                        name
                    }
                    dataset {
                        id
                        slug
                        name
                        description
                        latestVersion {
                            id
                            name
                            description
                        }
                    }
                }
            }
        """,
            {"datasetSlug": identifier, "workspaceSlug": self.slug},
        )
        data = response["datasetLinkBySlug"]

        if data is None:
            raise ValueError(f"Dataset {identifier} does not exist.")

        return Dataset(
            id=data["dataset"]["id"],
            slug=data["dataset"]["slug"],
            name=data["dataset"]["name"],
            description=data["dataset"]["description"],
        )

    def list_datasets(self) -> list[Dataset]:
        """List datasets in a workspace.

        Returns
        -------
        List of Datasets
        """
        response = graphql(
            """
            query getWorkspaceDatasets($slug: String!) {
                workspace(slug: $slug) {
                    datasets {
                        items { 
                            id
                            dataset {
                                id
                                slug
                                name
                                description
                            }
                        }
                    
                    }
                }
            }
        """,
            {"slug": self.slug},
        )
        data = response["workspace"]["datasets"]["items"]
        datasets = [
            Dataset(
                id=d["dataset"]["id"],
                slug=d["dataset"]["slug"],
                name=d["dataset"]["name"],
                description=d["dataset"]["description"],
            )
            for d in data
        ]

        return datasets



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List, Literal, Optional

from pydantic import Field

from .base_model import BaseModel
from .enums import CreateConnectionError


class CreateConnection(BaseModel):
    create_connection: "CreateConnectionCreateConnection" = Field(
        alias="createConnection"
    )


class CreateConnectionCreateConnection(BaseModel):
    success: bool
    connection: Optional["CreateConnectionCreateConnectionConnection"]
    errors: List[CreateConnectionError]


class CreateConnectionCreateConnectionConnection(BaseModel):
    typename__: Literal[
        "Connection",
        "CustomConnection",
        "DHIS2Connection",
        "GCSConnection",
        "IASOConnection",
        "PostgreSQLConnection",
        "S3Connection",
    ] = Field(alias="__typename")
    id: str
    name: str


CreateConnection.model_rebuild()
CreateConnectionCreateConnection.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List

from pydantic import Field

from .base_model import BaseModel
from .enums import UpdateWebappError


class UpdateWebapp(BaseModel):
    update_webapp: "UpdateWebappUpdateWebapp" = Field(alias="updateWebapp")


class UpdateWebappUpdateWebapp(BaseModel):
    success: bool
    errors: List[UpdateWebappError]


UpdateWebapp.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import List

from pydantic import Field

from .base_model import BaseModel
from .enums import PipelineRecipientError


class AddPipelineRecipient(BaseModel):
    add_pipeline_recipient: "AddPipelineRecipientAddPipelineRecipient" = Field(
        alias="addPipelineRecipient"
    )


class AddPipelineRecipientAddPipelineRecipient(BaseModel):
    success: bool
    errors: List[PipelineRecipientError]


AddPipelineRecipient.model_rebuild()



# Generated by ariadne-codegen
# Source: openhexa/graphql/queries.graphql

from typing import Any, List, Optional

from pydantic import Field

from .base_model import BaseModel
from .enums import PipelineRunStatus, PipelineType


class Pipeline(BaseModel):
    pipeline_by_code: Optional["PipelinePipelineByCode"] = Field(alias="pipelineByCode")


class PipelinePipelineByCode(BaseModel):
    id: Any
    code: str
    name: Optional[str]
    description: Optional[str]
    type: PipelineType
    current_version: Optional["PipelinePipelineByCodeCurrentVersion"] = Field(
        alias="currentVersion"
    )
    runs: "PipelinePipelineByCodeRuns"
    webhook_url: Optional[str] = Field(alias="webhookUrl")
    webhook_enabled: bool = Field(alias="webhookEnabled")
    schedule: Optional[str]
    notebook_path: Optional[str] = Field(alias="notebookPath")
    source_template: Optional["PipelinePipelineByCodeSourceTemplate"] = Field(
        alias="sourceTemplate"
    )
    has_new_template_versions: bool = Field(alias="hasNewTemplateVersions")
    new_template_versions: List["PipelinePipelineByCodeNewTemplateVersions"] = Field(
        alias="newTemplateVersions"
    )
    recipients: List["PipelinePipelineByCodeRecipients"]


class PipelinePipelineByCodeCurrentVersion(BaseModel):
    id: Any
    name: Optional[str]
    version_number: int = Field(alias="versionNumber")
    user: Optional["PipelinePipelineByCodeCurrentVersionUser"]
    created_at: Any = Field(alias="createdAt")


class PipelinePipelineByCodeCurrentVersionUser(BaseModel):
    id: Any
    email: str
    display_name: str = Field(alias="displayName")
    avatar: "PipelinePipelineByCodeCurrentVersionUserAvatar"


class PipelinePipelineByCodeCurrentVersionUserAvatar(BaseModel):
    initials: str
    color: str


class PipelinePipelineByCodeRuns(BaseModel):
    items: List["PipelinePipelineByCodeRunsItems"]


class PipelinePipelineByCodeRunsItems(BaseModel):
    id: Any
    status: PipelineRunStatus
    execution_date: Optional[Any] = Field(alias="executionDate")
    user: Optional["PipelinePipelineByCodeRunsItemsUser"]


class PipelinePipelineByCodeRunsItemsUser(BaseModel):
    id: Any
    display_name: str = Field(alias="displayName")
    email: str


class PipelinePipelineByCodeSourceTemplate(BaseModel):
    id: Any
    code: str
    name: str


class PipelinePipelineByCodeNewTemplateVersions(BaseModel):
    id: Any
    changelog: Optional[str]
    version_number: int = Field(alias="versionNumber")
    created_at: Any = Field(alias="createdAt")


class PipelinePipelineByCodeRecipients(BaseModel):
    user: "PipelinePipelineByCodeRecipientsUser"


class PipelinePipelineByCodeRecipientsUser(BaseModel):
    id: Any
    display_name: str = Field(alias="displayName")


Pipeline.model_rebuild()
PipelinePipelineByCode.model_rebuild()
PipelinePipelineByCodeCurrentVersion.model_rebuild()
PipelinePipelineByCodeCurrentVersionUser.model_rebuild()
PipelinePipelineByCodeRuns.model_rebuild()
PipelinePipelineByCodeRunsItems.model_rebuild()
PipelinePipelineByCodeRecipients.model_rebuild()
